import os
import shutil
import subprocess
import json
from kelpie import io
from kelpie.structure import Structure
from kelpie import files_and_folders
from kelpie.vasp_settings.incar import DEFAULT_VASP_INCAR_SETTINGS
from kelpie.vasp_input_generator import VaspInputGenerator
from kelpie.vasp_calculation_data import VaspCalculationData


class KelpieWorkflowError(Exception):
    """Base class to handle calculation workflow related errors."""
    pass


class GenericWorkflow(object):
    """Generic workflow class."""

    def __init__(self,
                 initial_structure=None,
                 run_location=None,
                 custom_calculation_settings=None,
                 mpi_call=None,
                 **kwargs):
        """Constructor.

        :param initial_structure: `kelpie.structure` with the initial structure
        :param run_location: String with the location where calculations should be performed.
        :param custom_calculation_settings: Dictionary of *nondefault* calculation settings
        :param mpi_call: String with the complete MPI call to use for calculations.
        :param kwargs: Other miscellaneous arguments.
        """
        self._initial_structure = None
        self.initial_structure = initial_structure

        self._run_location = None
        self.run_location = run_location

        self._custom_calculation_settings = None
        self.custom_calculation_settings = custom_calculation_settings

        self._mpi_call = None
        self.mpi_call = mpi_call

        self.kwargs = kwargs

    @property
    def initial_structure(self):
        return self._initial_structure

    @initial_structure.setter
    def initial_structure(self, initial_structure):
        if not isinstance(initial_structure, Structure):
            error_message = 'Arugment `initial_structure` must be a `kelpie.structure.Structure object'
            raise KelpieWorkflowError(error_message)
        self._initial_structure = initial_structure

    @property
    def run_location(self):
        return self._run_location

    @run_location.setter
    def run_location(self, run_location):
        if not run_location:
            error_message = 'Run location not specified'
            raise KelpieWorkflowError(error_message)
        if not os.path.isdir(run_location):
            error_message = 'Specified run location {} not found'.format(run_location)
            raise KelpieWorkflowError(error_message)
        self._run_location = run_location

    @property
    def custom_calculation_settings(self):
        return self._custom_calculation_settings

    @custom_calculation_settings.setter
    def custom_calculation_settings(self, custom_calculation_settings):
        if not custom_calculation_settings:
            self._custom_calculation_settings = {}
        else:
            self._custom_calculation_settings = custom_calculation_settings

    @property
    def mpi_call(self):
        return self._mpi_call

    @mpi_call.setter
    def mpi_call(self, mpi_call):
        if not mpi_call:
            error_message = 'MPI call for the calculation not specified'
            raise KelpieWorkflowError(error_message)
        self._mpi_call = mpi_call

    @staticmethod
    def run_vasp(mpi_call):
        with open('stdout.txt', 'w') as fstdout, open('stderr.txt', 'w') as fstderr:
            vasp_process = subprocess.run(mpi_call.split(), stdout=fstdout, stderr=fstderr)
        return vasp_process

    def do_relaxation(self, structure=None, settings=None, mpi_call=None, **kwargs):
        # propagate variable "n_attempts" to keep track of the number of VASP runs
        # if this is the first do_relaxation() call, initialize "n_attempts"
        if not hasattr(kwargs, 'n_attempts'):
            kwargs['n_attempts'] = 0

        # write input files with the given structure and settings
        ig = VaspInputGenerator(structure=structure,
                                calculation_settings=settings,
                                **kwargs)
        ig.write_vasp_input_files()
        # use the MPI call specified to run VASP
        vasp_process = self.run_vasp(mpi_call)
        # did the VASP run OK?
        if vasp_process.returncode != 0:
            error_message = 'Something went wrong with the MPI VASP run'
            raise KelpieWorkflowError(error_message)
        # increase "n_attempts": VASP has already been run once
        kwargs['n_attempts'] += 1
        # parse the calculation output and check if relevent convergence criteria are met
        vcd = VaspCalculationData(vasprunxml_file='vasprun.xml')
        converged = vcd.is_fully_converged(scf_thresh=settings.get('ediff'),
                                           force_thresh=settings.get('ediffg'))

        # recursively call do_relaxation() until either convergence or maximum attempts have been reached
        if not converged and kwargs['n_attempts'] <= 5:
            # if not converged wrt number of bands, restart with
            # 1.2*N_bands_old or N_bands_old+4 whichever is higher
            # make sure settings are carried over to the static calculation
            if not vcd.is_number_of_bands_converged():
                n_bands = max([int(vcd.n_bands * 1.2), vcd.n_bands + 4])
                settings.update({'n_bands': n_bands})
                self._custom_calculation_settings.update({
                    'relaxation': {'n_bands': n_bands},
                    'acc_std_relax': {'n_bands': n_bands},
                    'static': {'n_bands': n_bands}
                })
            try:
                output_structure = io.read_poscar('CONTCAR')
            except io.KelpieIOError:
                error_message = 'Could not read the CONTCAR file'
                raise KelpieWorkflowError(error_message)
            else:
                files_and_folders.backup_files()
                return self.do_relaxation(structure=output_structure,
                                          settings=settings,
                                          mpi_call=mpi_call,
                                          **kwargs)
        else:
            return vcd, converged

    def do_static(self, structure=None, settings=None, mpi_call=None, **kwargs):
        # propagate variable "n_attempts" to keep track of the number of VASP runs
        # if this is the first do_relaxation() call, initialize "n_attempts"
        if not hasattr(kwargs, 'n_attempts'):
            kwargs['n_attempts'] = 0

        # write input files with the given structure and settings
        ig = VaspInputGenerator(structure=structure,
                                calculation_settings=settings,
                                **kwargs)
        ig.write_vasp_input_files()
        # use the MPI call specified to run VASP
        vasp_process = self.run_vasp(mpi_call)
        # did the VASP run OK?
        if vasp_process.returncode != 0:
            error_message = 'Something went wrong with the MPI VASP run'
            raise KelpieWorkflowError(error_message)
        # increase "n_attempts": VASP has already been run once
        kwargs['n_attempts'] += 1
        # parse the calculation output and check if relevent convergence criteria are met
        vcd = VaspCalculationData(vasprunxml_file='vasprun.xml',
                                  vasp_outcar_file='OUTCAR')
        converged = vcd.is_scf_converged(threshold=settings.get('ediff'))

        # recursively call do_static() until either convergence or maximum attempts have been reached
        max_nattempts = kwargs.get('max_nattempts', 1)
        if not converged and kwargs['n_attempts'] < max_nattempts:
            files_and_folders.backup_files()
            return self.do_static(structure=structure,
                                  settings=settings,
                                  mpi_call=mpi_call,
                                  **kwargs)
        else:
            return vcd, converged


class RelaxationWorkflow(GenericWorkflow):
    """Class with workflow for a relaxation run followed by a final static run."""

    def __init__(self,
                 initial_structure=None,
                 run_location=None,
                 custom_calculation_settings=None,
                 mpi_call=None,
                 **kwargs):
        super(RelaxationWorkflow, self).__init__(initial_structure=initial_structure,
                                                 run_location=run_location,
                                                 custom_calculation_settings=custom_calculation_settings,
                                                 mpi_call=mpi_call,
                                                 **kwargs)

    def perform_workflow(self, from_scratch=False):
        relaxation_dir = os.path.join(self.run_location, 'relaxation')
        # if from_scratch, delete the "relaxation" folder, if it exists
        if from_scratch and os.path.isdir(relaxation_dir):
            shutil.rmtree(relaxation_dir)
        # create a "relaxation" folder, if one doesn't already exist
        os.makedirs(relaxation_dir, exist_ok=True)
        relaxation_settings = DEFAULT_VASP_INCAR_SETTINGS['relaxation']
        relaxation_settings.update(self.custom_calculation_settings.get('relaxation', {}))
        initial_structure = self.initial_structure
        with files_and_folders.change_working_dir(relaxation_dir):
            if not from_scratch:
                previous_outcar = os.path.join(relaxation_dir, 'OUTCAR')
                previous_contcar = os.path.join(relaxation_dir, 'CONTCAR')
                previous_poscar = os.path.join(relaxation_dir, 'POSCAR')
                if os.path.isfile(previous_outcar) and os.path.getsize(previous_outcar):
                    files_and_folders.backup_files()
                if os.path.isfile(previous_contcar) and os.path.getsize(previous_contcar):
                    initial_structure = io.read_poscar(previous_contcar)
                elif os.path.isfile(previous_poscar) and os.path.getsize(previous_poscar):
                    initial_structure = io.read_poscar(previous_poscar)
            vcd, converged = self.do_relaxation(structure=initial_structure,
                                                settings=relaxation_settings,
                                                mpi_call=self.mpi_call,
                                                **self.kwargs)
        with open('relaxation_data.json', 'w') as fw:
            json.dump(vcd.as_dict(), fw, indent=2)

        if not converged:
            error_message = 'Error while performing the relaxation run(s)'
            raise KelpieWorkflowError(error_message)

        relaxation_output_structure = os.path.join(relaxation_dir, 'CONTCAR')
        initial_structure = io.read_poscar(relaxation_output_structure)
        static_dir = os.path.join(self.run_location, 'static')
        os.makedirs(static_dir, exist_ok=True)
        static_settings = DEFAULT_VASP_INCAR_SETTINGS['static']
        static_settings.update(self.custom_calculation_settings.get('static', {}))
        files_and_folders.copy_files(src_folder='relaxation',
                                     dest_folder='static',
                                     list_of_filenames=['CHGCAR'])
        with files_and_folders.change_working_dir(static_dir):
            vcd, converged = self.do_static(structure=initial_structure,
                                            settings=static_settings,
                                            mpi_call=self.mpi_call,
                                            **self.kwargs)
        with open('static_data.json', 'w') as fw:
            json.dump(vcd.as_dict(), fw, indent=2)

        if not converged:
            error_message = 'Error while performing the final static run'
            raise KelpieWorkflowError(error_message)


class AccStdRelaxWorkflow(GenericWorkflow):
    """Class with workflow for an accurate (e.g. for phonons) relaxation of the
    Spglib-standardized primitive unit cell."""

    def __init__(self,
                 initial_structure=None,
                 run_location=None,
                 custom_calculation_settings=None,
                 mpi_call=None,
                 **kwargs):
        super(AccStdRelaxWorkflow, self).__init__(initial_structure=initial_structure,
                                                  run_location=run_location,
                                                  custom_calculation_settings=custom_calculation_settings,
                                                  mpi_call=mpi_call,
                                                  **kwargs)

    def perform_workflow(self, from_scratch=False):
        #TODO: Make the directory structure more organic (avoid duplicate
        # "initial_structure.vasp" files e.g.)
        ##relaxation_dir = os.path.join(self.run_location, 'acc_std_relax')
        relaxation_dir = os.path.join(self.run_location)
        ### if from_scratch, delete the relaxation folder, if it exists
        ##if from_scratch and os.path.isdir(relaxation_dir):
        ##    shutil.rmtree(relaxation_dir)
        ### create a "relaxation" folder, if one doesn't already exist
        ##os.makedirs(relaxation_dir, exist_ok=True)
        relaxation_settings = DEFAULT_VASP_INCAR_SETTINGS['acc_std_relax']
        relaxation_settings.update(self.custom_calculation_settings.get('acc_std_relax', {}))
        initial_structure = self.initial_structure
        with files_and_folders.change_working_dir(relaxation_dir):
            if not from_scratch:
                previous_outcar = os.path.join(relaxation_dir, 'OUTCAR')
                previous_contcar = os.path.join(relaxation_dir, 'CONTCAR')
                previous_poscar = os.path.join(relaxation_dir, 'POSCAR')
                if os.path.isfile(previous_outcar) and os.path.getsize(previous_outcar):
                    files_and_folders.backup_files()
                if os.path.isfile(previous_contcar) and os.path.getsize(previous_contcar):
                    initial_structure = io.read_poscar(previous_contcar)
                elif os.path.isfile(previous_poscar) and os.path.getsize(previous_poscar):
                    initial_structure = io.read_poscar(previous_poscar)
            vcd, converged = self.do_relaxation(structure=initial_structure,
                                                settings=relaxation_settings,
                                                mpi_call=self.mpi_call,
                                                **self.kwargs)
        with open('acc_std_relax_data.json', 'w') as fw:
            json.dump(vcd.as_dict(), fw, indent=2)

        if not converged:
            error_message = 'Error while performing the relaxation run(s)'
            raise KelpieWorkflowError(error_message)


class ScForcesWorkflow(GenericWorkflow):
    """Class with the workflow for the calculation of forces in supercells with
    (a) displaced atom(s)."""

    def __init__(self,
                 initial_structure=None,
                 run_location=None,
                 custom_calculation_settings=None,
                 mpi_call=None,
                 **kwargs):
        super(ScForcesWorkflow, self).__init__(initial_structure=initial_structure,
                                               run_location=run_location,
                                               custom_calculation_settings=custom_calculation_settings,
                                               mpi_call=mpi_call,
                                               **kwargs)

    def perform_workflow(self, from_scratch=False):
        calc_dir = os.path.join(self.run_location)
        calc_settings = DEFAULT_VASP_INCAR_SETTINGS['sc_forces']
        calc_settings.update(self.custom_calculation_settings.get('sc_forces', {}))
        initial_structure = self.initial_structure
        with files_and_folders.change_working_dir(calc_dir):
            if not from_scratch:
                previous_outcar = os.path.join(calc_dir, 'OUTCAR')
                if os.path.isfile(previous_outcar) and os.path.getsize(previous_outcar):
                    files_and_folders.backup_files()
            vcd, converged = self.do_static(structure=initial_structure,
                                            settings=calc_settings,
                                            mpi_call=self.mpi_call,
                                            **self.kwargs)
        with open('sc_forces_data.json', 'w') as fw:
            json.dump(vcd.as_dict(), fw, indent=2)

        if not converged:
            error_message = 'Error during (scf) calculation of forces'
            raise KelpieWorkflowError(error_message)




import json
import os

# GLOBAL data

# Standard atomic weights
# Source: Meija et al., "Atomic weights of the elements 2013 (IUPAC Tehcnical Report)", Pure Appl. Chem. 88, 265 (2016)
# For elements missing from the above list, source:
# Sansonetti et al., Handbook of Basic Atomic Spectroscopic Data (version 1.1.3). NIST, Gaithersburg, MD
# Accessed from http://physics.nist.gov/Handbook [Date: Dec 27, 2017]
atomic_weights_file = os.path.join(os.path.dirname(__file__), 'standard_atomic_weights.json')
with open(atomic_weights_file, 'r') as fr:
    STD_ATOMIC_WEIGHTS = json.load(fr)


# Number of electrons in s, p, d, f shells
# Source: OQMD (for which the data was taken from Mathematica, IIRC)
valence_electrons_file = os.path.join(os.path.dirname(__file__), 'valence_electrons.json')
with open(valence_electrons_file, 'r') as fr:
    VALENCE_ELECTRONS = json.load(fr)



import os
import shutil
import subprocess
import json
import six
from kelpie import files_and_folders
from kelpie.scheduler_settings import DEFAULT_SCHEDULER_SETTINGS
from kelpie.scheduler_templates import SCHEDULER_TEMPLATES


class KelpieBreederError(Exception):
    """Base class to handle errors associated with breeding."""
    pass


class KelpieBreeder(object):
    """Base class to create the run directory and write batch script."""

    def __init__(self,
                 input_structure_file=None,
                 run_location=None,
                 host_scheduler_settings=None,
                 custom_scheduler_settings=None,
                 batch_script_template=None,
                 submit_batch_job=None,
                 **kwargs):
        """Constructor.

        :param input_structure_file: String with the location of the VASP5 POSCAR with the initial structure.
        :param run_location: String with the location where VASP calculations should be performed.
                             (Default: location of the specified `input_structure_file`)
        :param host_scheduler_settings: String with the name of a predefined host or path to a JSON file with the
                                        scheduler settings. Must be one of the hosts defined in
                                        `kelpie.scheduler_settings` (with a corresponding
                                        "[host_scheduler_tag].json" file) OR path to a JSON file with settings.
                                        Note: Path to a JSON file takes precedence.
                                        E.g. "cori_knl", "cori_haswell" (predefined) OR "/path/to/settings.json"
                                        (Default: "cori_knl")
        :param custom_scheduler_settings: Dictionary of tags, and *nondefault* values to go into the batch script.
                                          (The default settings dictionary, loaded for the specified
                                          `host_scheduler_settings` above will be updated)
                                          Some common tags:
                                          - job-name
                                          - partition
                                          - qos
                                          - account
                                          - constraint
                                          - license
                                          - nodes
                                          - walltime
                                          - output
                                          - omp_threads
                                          - modules (to load)
                                          - executable
                                          - n_mpi_per_node (# MPI tasks to invoke per node)
                                          - n_cpu_bind (# logical CPUs to bind to an MPI task)
        :param batch_script_template: String with the path to a batch script template file OR the name of a
                                      predefined template in the "scheduler_settings" directory.
                                      Note: Path to a file takes precedence.
                                      (Default: "cori")
        :param submit_batch_job: Boolean specifying whether the batch job should be submitted to the scheduler.
                                 (Default: True)
        :param kwargs: Dictionary of other miscellaneous parameters, if any.
        """

        #: VASP POSCAR file containing the structure (only VASP 5 format currently supported).
        #: `kelpie.structure.Structure` object containing VASP POSCAR data.
        self._input_structure_file = None
        self.input_structure_file = input_structure_file

        #: Relative/absolute path to run VASP calculations.
        self._run_location = None
        self.run_location = run_location

        #: Dictionary of batch scheduler settings corresponding to the host where the calculations are being run. Can
        #  be path a JSON file with the settings OR one of the predefined tags in
        # `kelpie.scheduler_settings.DEFAULT_SCHEDULER_SETTINGS`. File takes precedence.
        self._host_scheduler_settings = None
        self.host_scheduler_settings = host_scheduler_settings

        #: Nondefault scheduler settings for this particular run. Updates `self.host_scheduler_settings`.
        self._custom_scheduler_settings = None
        self.custom_scheduler_settings = custom_scheduler_settings

        #: Template for the batch script
        self._batch_script_template = None
        self.batch_script_template = batch_script_template

        #: Should the batch job be submitted to the scheduler?
        self._submit_batch_job = None
        self.submit_batch_job = submit_batch_job

        #: Unsupported keyword arguments
        self.kwargs = kwargs

    @property
    def input_structure_file(self):
        return self._input_structure_file

    @input_structure_file.setter
    def input_structure_file(self, input_structure_file):
        if not input_structure_file:
            error_message = 'Input structure file not specified'
            raise KelpieBreederError(error_message)
        if not os.path.isfile(input_structure_file):
            error_message = 'Specified `input_structure_file` {} not found'.format(input_structure_file)
            raise KelpieBreederError(error_message)
        self._input_structure_file = os.path.abspath(input_structure_file)

    @property
    def run_location(self):
        return self._run_location

    @run_location.setter
    def run_location(self, run_location):
        if not run_location:
            self._run_location = os.path.dirname(os.path.abspath(self.input_structure_file))
        else:
            self._run_location = run_location

    @property
    def host_scheduler_settings(self):
        return self._host_scheduler_settings

    @host_scheduler_settings.setter
    def host_scheduler_settings(self, host_scheduler_settings):
        if not host_scheduler_settings:
            self._host_scheduler_settings = DEFAULT_SCHEDULER_SETTINGS['cori_knl']
            return
        if os.path.isfile(host_scheduler_settings):
            with open(host_scheduler_settings, 'r') as fr:
                self._host_scheduler_settings = json.load(fr)
        else:
            settings = DEFAULT_SCHEDULER_SETTINGS.get(host_scheduler_settings)
            if not settings:
                error_message = 'Unable to find the specified `host_scheduler_settings`'
                raise KelpieBreederError(error_message)
            self._host_scheduler_settings = settings

    @property
    def custom_scheduler_settings(self):
        return self._custom_scheduler_settings

    @custom_scheduler_settings.setter
    def custom_scheduler_settings(self, custom_scheduler_settings):
        if not custom_scheduler_settings:
            self._custom_scheduler_settings = {}
            return
        if os.path.isfile(custom_scheduler_settings):
            with open(custom_scheduler_settings, 'r') as fr:
                self._custom_scheduler_settings = json.load(fr)
        else:
            self._custom_scheduler_settings = custom_scheduler_settings

    @property
    def batch_script_template(self):
        return self._batch_script_template

    @batch_script_template.setter
    def batch_script_template(self, batch_script_template):
        if not batch_script_template:
            self._batch_script_template = SCHEDULER_TEMPLATES['cori']
            return
        if os.path.isfile(batch_script_template):
            self._batch_script_template = batch_script_template
        else:
            template = SCHEDULER_TEMPLATES.get(batch_script_template)
            if not template:
                error_message = 'Unable to find the specified `batch_script_template`'
                raise KelpieBreederError(error_message)
            self._batch_script_template = template

    @property
    def submit_batch_job(self):
        return self._submit_batch_job

    @submit_batch_job.setter
    def submit_batch_job(self, submit_batch_job):
        if not submit_batch_job:
            self._submit_batch_job = True
            return
        if isinstance(submit_batch_job, six.string_types):
            self._submit_batch_job = submit_batch_job.lower()[0] == 't'
        else:
            self._submit_batch_job = submit_batch_job

    @property
    def scheduler_script_name(self):
        return '{}.q'.format(os.path.splitext(os.path.basename(self.batch_script_template))[0])

    def _get_batch_script(self):
        # general batch scheduler settings
        settings = {**self.host_scheduler_settings}
        settings.update(self.custom_scheduler_settings)

        # generate the "module load" section
        modules = []
        for module in settings.get('modules', []):
            modules.append('module load {}'.format(module))
        modules = '\n'.join(modules)
        settings.update({'modules': modules})

        # if calculation workflow and custom calculation settings were specified,
        # write them as is into the batch script
        calculation_params = []
        calculation_workflow = self.kwargs.get('calculation_workflow')
        if calculation_workflow:
            calculation_params.append('-w {}'.format(calculation_workflow))
        custom_calculation_settings = self.kwargs.get('custom_calculation_settings')
        if custom_calculation_settings:
            calculation_params.append('-ccs {}'.format(custom_calculation_settings))
        calculation_params = ' '.join(calculation_params)

        # arguments for kelpie grazer
        settings.update({'run_location': self.run_location,
                         'calculation_params': calculation_params})

        # format the template with all the arguments
        with open(self.batch_script_template, 'r') as fr:
            template = fr.read()
        template = template.format(**settings, **self.kwargs)
        return template

    @property
    def batch_script(self):
        return self._get_batch_script()

    def submit_job_to_scheduler(self):
        settings = {**self.host_scheduler_settings}
        settings.update(self.custom_scheduler_settings)
        submit_cmd = settings.get('submit_cmd')
        if not submit_cmd:
            error_message = 'Submit command for the batch scheduler not specified'
            raise KelpieBreederError(error_message)
        with files_and_folders.change_working_dir(self.run_location):
            subprocess.run([submit_cmd, self.scheduler_script_name], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    def _get_mpi_call(self):
        settings = {**self.host_scheduler_settings}
        settings.update(self.custom_scheduler_settings)
        mpi_call = settings.get('mpi_call')
        if not mpi_call:
            error_message = 'MPI call not specified for the host (in scheduler settings)'
            raise KelpieBreederError(error_message)
        # special settings for the "shared" queue on Cori
        if not settings.get('nodes'):
            settings.update({'n_mpi': int(settings.get('cores'))})
        else:
            settings.update({'n_mpi': int(settings.get('nodes'))*int(settings.get('n_mpi_per_node'))})
        mpi_call = mpi_call.format(**settings)
        return mpi_call

    @property
    def mpi_call(self):
        return self._get_mpi_call()

    def breed(self):
        """Create the directory for running calculations if not already present, copy the specified input structure,
        write the batch script and the mpi_call command into the run location, submit the batch job.
        """
        # create the directory structure if not already present
        if not os.path.isdir(self.run_location):
            os.makedirs(self.run_location)

        # copy the specified input structure file into "[self.run_location]/initial_structure"
        init_structure_file = os.path.join(self.run_location, 'initial_structure.vasp')
        shutil.copy(self.input_structure_file, init_structure_file)

        # write the mpi call into "[self.run_location]/mpi_call.txt". Will be read while grazing
        mpi_call_file = os.path.join(self.run_location, 'mpi_call.txt')
        with open(mpi_call_file, 'w') as fw:
            fw.write(self.mpi_call)

        # write the batch script
        batch_script_file = os.path.join(self.run_location, self.scheduler_script_name)
        with open(batch_script_file, 'w') as fw:
            fw.write(self.batch_script)

        if self.submit_batch_job:
            self.submit_job_to_scheduler()





import os
import numpy as np
import kelpie.vasp_output_parser as vasp_output_parser


class DOSError(Exception):
    """Base class to handle error(s) related to density of states."""
    pass


class DOS(object):
    """Base class to store density of states data from a calculation."""

    def __init__(self, vasprunxml_file=None, doscar_file=None):
        """
        :param vasprunxml_file: String with the path to the vasprun.xml file from the calculation.
                                (takes preference over doscar_file)
        :param doscar_file: String with the path to the DOSCAR file from the calculation.
        """
        self._vasprunxml_file = None
        self._vxparser = None
        self.vasprunxml_file = vasprunxml_file

        self._doscar_file = None
        self.doscar_file = doscar_file
        self._doscarparser = None

        self._total_dos_data = self.get_total_dos_data()
        self._fermi_energy = self.get_fermi_energy()
        self._energies = self.get_energies()
        self._total_dos = self.get_total_dos()
        self._total_integrated_dos = self.get_total_integrated_dos()
        self._projected_dos = None

    @property
    def vasprunxml_file(self):
        return self._vasprunxml_file

    @vasprunxml_file.setter
    def vasprunxml_file(self, vasprunxml_file):
        if vasprunxml_file:
            if os.path.isfile(vasprunxml_file):
                self._vasprunxml_file = vasprunxml_file
                self._vxparser = vasp_output_parser.VasprunXMLParser(vasprunxml_file=vasprunxml_file)
            else:
                error_message = 'Cannot find the vasprun.xml file specified: {}'.format(vasprunxml_file)
                raise DOSError(error_message)

    @property
    def vxparser(self):
        return self._vxparser

    @property
    def doscar_file(self):
        return self._doscar_file

    @doscar_file.setter
    def doscar_file(self, doscar_file):
        if doscar_file:
            if os.path.isfile(doscar_file):
                self._doscar_file = doscar_file
            else:
                error_message = 'Cannot find the DOSCAR file specified: {}'.format(doscar_file)
                raise DOSError(error_message)

    @property
    def total_dos_data(self):
        return self._total_dos_data

    def get_total_dos_data(self):
        if self.vxparser is None:
            return
        return self.vxparser.read_total_density_of_states()

    @property
    def fermi_energy(self):
        return self._fermi_energy

    def get_fermi_energy(self):
        if self.vasprunxml_file is not None:
            return self.get_fermi_energy_from_xml()
        elif self.doscar_file is not None:
            return self.get_fermi_energy_from_doscar()

    def get_fermi_energy_from_xml(self):
        return self.vxparser.read_fermi_energy()

    def get_fermi_energy_from_doscar(self):
        raise NotImplementedError

    @property
    def energies(self):
        return self._energies

    def get_energies(self):
        if self.vasprunxml_file is not None:
            return self.get_energies_from_xml()
        elif self.doscar_file is not None:
            return self.get_energies_from_doscar()

    def get_energies_from_xml(self):
        if not self.total_dos_data:
            return
        return np.array([gridpoint[0] - self.fermi_energy for gridpoint in self.total_dos_data['spin_1']])

    def get_energies_from_doscar(self):
        raise NotImplementedError

    @property
    def total_dos(self):
        return self._total_dos

    def get_total_dos(self):
        if self.vasprunxml_file is not None:
            return self.get_total_dos_from_xml()
        elif self.doscar_file is not None:
            return self.get_total_dos_from_doscar()

    def get_total_dos_from_xml(self):
        if not self.total_dos_data:
            return
        total_dos = {}
        for spin in self.total_dos_data:
            total_dos[spin] = np.array([gridpoint[1] for gridpoint in self.total_dos_data[spin]])
        return total_dos

    def get_total_dos_from_doscar(self):
        raise NotImplementedError

    @property
    def total_integrated_dos(self):
        return self._total_integrated_dos

    def get_total_integrated_dos(self):
        if self.vasprunxml_file is not None:
            return self.get_total_integrated_dos_from_xml()
        elif self.doscar_file is not None:
            return self.get_total_integrated_dos_from_doscar()

    def get_total_integrated_dos_from_xml(self):
        if not self.total_dos_data:
            return
        total_intdos = {}
        for spin in self.total_dos_data:
            total_intdos[spin] = np.array([gridpoint[2] for gridpoint in self.total_dos_data[spin]])
        return total_intdos

    def get_total_integrated_dos_from_doscar(self):
        raise NotImplementedError

    @property
    def projected_dos(self):
        return self._projected_dos

    def get_projected_dos(self):
        if self.vasprunxml_file is not None:
            return self.get_projected_dos_from_xml()
        elif self.doscar_file is not None:
            return self.get_projected_dos_from_doscar()

    def get_projected_dos_from_xml(self):
        raise NotImplementedError

    def get_projected_dos_from_doscar(self):
        raise NotImplementedError

    def _check_if_metal(self, spin='spin_1', tol=1e-3):
        if self.energies is None or not self.energies.size:
            return
        efermi_index = np.abs(self.energies).argmin()
        if self.energies[efermi_index] < 0:
            bracket_index = efermi_index + 1
        else:
            bracket_index = efermi_index - 1
        return (self.total_dos[spin][efermi_index] > tol) and (self.total_dos[spin][bracket_index] > tol)

    def check_if_metal(self, tol=1e-3):
        return dict([(spin, self._check_if_metal(spin=spin, tol=tol)) for spin in self.total_dos])

    @property
    def is_metal(self):
        return self.check_if_metal()

    def _find_vbm(self, spin='spin_1', tol=1e-3):
        is_metal = self._check_if_metal(spin=spin, tol=tol)
        if is_metal is None or is_metal:
            return
        efermi_index = np.abs(self.energies).argmin()
        if self.energies[efermi_index] < 0:
            efermi_index += 1
        index = efermi_index
        while self.total_dos[spin][index] < tol:
            index -= 1
        return self.energies[index]

    def find_vbm(self, tol=1e-3):
        return dict([(spin, self._find_vbm(spin=spin, tol=tol)) for spin in self.total_dos])

    @property
    def vbm(self):
        return self.find_vbm()

    def _find_cbm(self, spin='spin_1', tol=1e-3):
        is_metal = self._check_if_metal(spin=spin, tol=tol)
        if is_metal is None or is_metal:
            return
        efermi_index = np.abs(self.energies).argmin()
        if self.energies[efermi_index] > 0:
            efermi_index -= 1
        index = efermi_index
        while self.total_dos[spin][index] < tol:
            index += 1
        return self.energies[index]

    def find_cbm(self, tol=1e-3):
        return dict([(spin, self._find_cbm(spin=spin, tol=tol)) for spin in self.total_dos])

    @property
    def cbm(self):
        return self.find_cbm()

    def _calculate_band_gap(self, spin='spin_1', tol=1e-3):
        is_metal = self._check_if_metal(spin=spin, tol=tol)
        if is_metal is not None:
            return (self.cbm[spin] - self.vbm[spin]) if not is_metal else 0.0

    def calculate_band_gap(self, tol=1e-3):
        return dict([(spin, self._calculate_band_gap(spin=spin, tol=tol)) for spin in self.total_dos])

    @property
    def band_gap(self):
        return self.calculate_band_gap()




import unittest
from kelpie.structure import Atom, AtomError
from kelpie.structure import Structure, StructureError


class TestAtom(unittest.TestCase):
    """Base class to test kelpie.structure.Atom class"""

    def test_no_coordinate(self):
        with self.assertRaises(AtomError):
            Atom(species='Al')

    def test_non_float_coordinates(self):
        with self.assertRaises(AtomError):
            Atom(coordinates=[0., 'A', 0.], species='Al')

    def test_coordinates_shape(self):
        with self.assertRaises(AtomError):
            Atom(coordinates=[], species='Al')
        with self.assertRaises(AtomError):
            Atom(coordinates=[0, 1, 2, 3], species='Al')

    def test_no_species(self):
        with self.assertRaises(AtomError):
            Atom(coordinates=[0, 0, 0])

    def test_species_starts_with_unknown_element_symbol(self):
        with self.assertRaises(AtomError):
            Atom(coordinates=[0, 0, 0], species='Af')

    def test_non_string_species(self):
        with self.assertRaises(AtomError):
            Atom(coordinates=[0, 0, 0], species='1Al')

    def test_all_OK(self):
        atom = Atom(coordinates=[0, 0.5, 0.989], species='Mn345')
        self.assertAlmostEqual(atom.x, 0.)
        self.assertAlmostEqual(atom.y, 0.5)
        self.assertAlmostEqual(atom.z, 0.989)
        self.assertEqual(atom.species, 'Mn345')
        self.assertEqual(atom.element, 'Mn345')


class TestStructure(unittest.TestCase):
    """Base class to test kelpie.structure.Structure class"""

    def setUp(self):
        self.structure = Structure()

    def test_default_scaling_constant(self):
        self.assertAlmostEqual(self.structure.scaling_constant, 1.0)

    def test_non_float_scaling_constant(self):
        with self.assertRaises(StructureError):
            self.structure.scaling_constant = 'c'

    def test_no_lattice_vectors(self):
        self.assertEqual(self.structure.lattice_vectors, None)

    def test_non_float_lattice_vectors(self):
        with self.assertRaises(StructureError):
            self.structure.lattice_vectors = [['a']]

    def test_shape_of_lattice_vectors(self):
        with self.assertRaises(StructureError):
            self.structure.lattice_vectors = [[1, 0, 0], [0, 1, 0], [0, 1]]

    def test_nan_lattice_vector_component(self):
        with self.assertRaises(StructureError):
            self.structure.lattice_vectors = [[1, 0, 0], [0, 1, 0], [0, None, 1]]
        with self.assertRaises(StructureError):
            self.structure.lattice_vectors = [[1, 0, 0], [float('nan'), 1, 0], [0, 0, 1]]

    def test_no_coordinate_system(self):
        self.assertEqual(self.structure.coordinate_system, None)

    def test_unrecognized_coordinate_system(self):
        with self.assertRaises(StructureError):
            self.structure.coordinate_system = 'Fractionl'

    def test_default_comment(self):
        self.assertEqual(self.structure.comment, 'Comment')

    def test_no_atoms(self):
        self.assertListEqual(self.structure.atoms, [])

    def test_atoms_instance(self):
        with self.assertRaises(StructureError):
            self.structure.atoms = ['Si']

    def test_add_atom(self):
        # non-`Atom` instance
        with self.assertRaises(StructureError):
            self.structure.add_atom('Si')
        # correct usage
        self.structure.add_atom(Atom([0, 0, 0], 'Si12'))
        self.assertEqual(len(self.structure.atoms), 1)
        self.assertListEqual(self.structure.list_of_species, ['Si12'])

    def test_all_OK(self):
        self.structure.comment = 'bcc-CsCl'
        self.structure.lattice_vectors = [[3.42, 0, 0], [0, 3.42, 0], [0, 0, 3.42]]
        self.structure.coordinate_system = 'direct'
        self.structure.add_atom(Atom([0, 0, 0], 'Cs'))
        self.structure.add_atom(Atom([0.5, 0.5, 0.5], 'Cl'))
        # composition_dict
        self.assertDictEqual(self.structure.composition_dict, {'Cs': 1, 'Cl': 1})
        # list_of_species
        self.assertListEqual(self.structure.list_of_species, ['Cl', 'Cs'])
        poscar = """bcc-CsCl
  1.00000000000000
  3.42000000000000    0.00000000000000    0.00000000000000
  0.00000000000000    3.42000000000000    0.00000000000000
  0.00000000000000    0.00000000000000    3.42000000000000
  Cl   Cs
   1    1
Direct
  0.50000000000000    0.50000000000000    0.50000000000000
  0.00000000000000    0.00000000000000    0.00000000000000"""
        self.assertEqual(self.structure.POSCAR, poscar)


if __name__ == '__main__':
    unittest.main()




import os
import unittest

sample_vasp_output_dir = os.path.join(os.path.dirname(__file__), 'sample_vasp_output')


class TestVasprunXMLParser(unittest.TestCase):
    """Base class to test kelpie.vasp_output_parser.VasprunXMLParser class."""

    def setUp(self):
        from kelpie import vasp_output_parser as parser
        self.vxparser = parser.VasprunXMLParser(os.path.join(sample_vasp_output_dir, 'relaxation_vasprun.xml.gz'))

    def test_read_composition_information(self):
        composition_info = self.vxparser.read_composition_information()
        self.assertTrue({'Ca', 'F'}.issubset(set(composition_info.keys())))
        self.assertAlmostEqual(composition_info['Ca']['atomic_mass'], 40.078)
        self.assertEqual(composition_info['F']['pseudopotential'], 'PAW_PBE F 08Apr2002')

    def test_read_list_of_atoms(self):
        self.assertListEqual(self.vxparser.read_list_of_atoms(), ['Ca', 'F', 'F'])

    def test_read_number_of_ionic_steps(self):
        self.assertEqual(self.vxparser.read_number_of_ionic_steps(), 3)

    def test_read_scf_energies(self):
        scf_energies = self.vxparser.read_scf_energies()
        self.assertIsInstance(scf_energies, dict)
        self.assertAlmostEqual(scf_energies[1][1], -17.7085613)

    def test_read_entropies(self):
        entropies = self.vxparser.read_entropies()
        self.assertIsInstance(entropies, dict)
        self.assertAlmostEqual(entropies[1], 0.01828511)

    def test_read_free_energies(self):
        energies = self.vxparser.read_free_energies()
        self.assertIsInstance(energies, dict)
        self.assertAlmostEqual(energies[2], -17.56163366)

    def test_read_forces(self):
        forces = self.vxparser.read_forces()
        self.assertEqual(len(forces.keys()), 3)
        self.assertListEqual(list(forces[0][2]), [0., 0., 0.])
        self.assertEqual(forces[1][1][0], 0.)

    def test_read_stress_tensors(self):
        stress_tensor = self.vxparser.read_stress_tensors()
        self.assertTupleEqual(stress_tensor[0].shape, (3, 3))
        self.assertAlmostEqual(stress_tensor[0][2][2], 70.96233651)

    def test_read_lattice_vectors(self):
        lattice_vectors = self.vxparser.read_lattice_vectors()
        self.assertTupleEqual(lattice_vectors[0].shape, (3, 3))
        self.assertAlmostEqual(lattice_vectors[1][1][1], 3.38391318)

    def test_read_atomic_coordinates(self):
        atomic_coordinates = self.vxparser.read_atomic_coordinates()
        self.assertEqual(len(atomic_coordinates[0]), 3)
        self.assertAlmostEqual(atomic_coordinates[1][1][2], 0.24999888)

    def test_read_kpoint_mesh(self):
        kpoint_mesh = self.vxparser.read_kpoint_mesh()
        self.assertListEqual(kpoint_mesh, [11, 12, 11])

    def test_read_irreducible_kpoints(self):
        irreducible_kpoints = self.vxparser.read_irreducible_kpoints()
        self.assertEqual(len(irreducible_kpoints), 402)
        self.assertEqual(len(irreducible_kpoints[-1]), 3)
        self.assertAlmostEqual(irreducible_kpoints[-1][1], -0.0833333)

    def test_read_cell_volumes(self):
        volume = self.vxparser.read_cell_volumes()
        self.assertAlmostEqual(volume[2], 41.78289124)

    def test_read_fermi_energy(self):
        self.assertAlmostEqual(self.vxparser.read_fermi_energy(), -1.11803743)

    def test_read_total_dos(self):
        total_dos = self.vxparser.read_total_density_of_states()
        self.assertListEqual(list(total_dos.keys()), ['spin_1', 'spin_2'])
        self.assertAlmostEqual(total_dos['spin_2'][12][1], -0.1063)

    def test_read_band_occupations(self):
        occupations = self.vxparser.read_band_occupations()
        self.assertIsInstance(occupations, dict)
        self.assertAlmostEqual(occupations['spin_1'][1]['band_energy'][5], -18.1784)
        self.assertAlmostEqual(occupations['spin_1'][1]['occupation'][9], 1.0353)
        self.assertEqual(max(occupations['spin_2'].keys()), 402)

    def test_read_run_timestamp(self):
        import datetime
        stamp = self.vxparser.read_run_timestamp()
        self.assertIsInstance(stamp, datetime.datetime)
        self.assertListEqual([stamp.year, stamp.month, stamp.day, stamp.hour, stamp.minute], [2016, 9, 22, 20, 31])

    def test_read_scf_looptimes(self):
        scf_looptimes = self.vxparser.read_scf_looptimes()
        self.assertEqual(len(scf_looptimes[0]), 15)
        self.assertListEqual(scf_looptimes[2][:3], [2.58, 2.47, 3.12])


if __name__ == '__main__':
    unittest.main()




import os
import json
from kelpie import io
from kelpie import calculation_workflows


class KelpieGrazerError(Exception):
    """Base class to handle errors associated with grazing."""
    pass


class KelpieGrazer(object):
    """Base class to perform specified calculation workflow(s)."""

    def __init__(self,
                 run_location=None,
                 initial_structure_file=None,
                 calculation_workflow=None,
                 custom_calculation_settings=None,
                 mpi_call_file=None,
                 **kwargs):
        """Constructor.

        :param run_location: String with the directory where calculations will be performed.
                             (Default: Current working directory)
        :param initial_structure_file: String with the location of the VASP5 POSCAR with the initial structure.
                                       (Default: "[self.run_location]/initial_structure.vasp")
        :param calculation_workflow: String with type of DFT calculation workflow.
                                     Currently, only "relaxation", "static" workflows implemented.
                                     (Default: "relaxation")
        :param custom_calculation_settings: Path to a JSON file with a dictionary of *nondefault* INCAR and POTCAR
                                            settings for each calculation type. The default calculation settings
                                            dictionary loaded according to the workflow will be updated.
                                            E.g., {"relaxation": {"ediff": 1E-8}, "static": {"sigma": 0.05}}
                                            (Default: {})
        :param mpi_call_file: Path to the text file with the mpi call command.
                              E.g. file contents: "srun -np 64 -c 4 -cpu_bind=cores vasp_std"
                              (Default: "[self.run_location]/mpi_call.txt"
        :param kwargs: Dictionary of other miscellaneous parameters, if any.
        """

        #: Relative/absolute path to run VASP calculations.
        self._run_location = None
        self.run_location = run_location

        #: File with the initial structure, and the initial structure
        self._initial_structure_file = None
        self.initial_structure_file = initial_structure_file
        self.initial_structure = io.read_poscar(self.initial_structure_file)

        #: Type of DFT calculation workflow: relaxation/static/hse/...
        self._calculation_workflow = None
        self.calculation_workflow = calculation_workflow
        #: Nondefault INCAR settings and POTCAR choices for different calculation types
        #: default INCAR, POTCAR settings defined by `kelpie.vasp_settings.incar.DEFAULT_VASP_INCAR_SETTINGS` for the
        # calculation types in the workflow specified.
        #: VASP recommended POTCARs used by default are in `kelpie.vasp_settings.potcar.VASP_RECO_POTCARS`.
        self._custom_calculation_settings = None
        self.custom_calculation_settings = custom_calculation_settings

        #: Path to the text file with the mpi call command
        self._mpi_call_file = None
        self.mpi_call_file = mpi_call_file

        #: Unsupported keyword arguments
        self.kwargs = kwargs

    @property
    def run_location(self):
        return self._run_location

    @run_location.setter
    def run_location(self, run_location):
        if not run_location:
            self._run_location = os.path.abspath(os.getcwd())
        else:
            self._run_location = os.path.abspath(run_location)

    @property
    def initial_structure_file(self):
        return self._initial_structure_file

    @initial_structure_file.setter
    def initial_structure_file(self, initial_structure_file):
        if not initial_structure_file:
            initial_structure_file = os.path.join(self.run_location, 'initial_structure.vasp')
        if not os.path.exists(initial_structure_file):
            error_message = 'Initial structure file {} not found'.format(initial_structure_file)
            raise KelpieGrazerError(error_message)
        else:
            self._initial_structure_file = os.path.abspath(initial_structure_file)

    @property
    def calculation_workflow(self):
        return self._calculation_workflow

    @calculation_workflow.setter
    def calculation_workflow(self, calculation_workflow):
        if not calculation_workflow:
            self._calculation_workflow = 'relaxation'
        elif calculation_workflow.lower() not in ['relaxation', 'static',
                                                  'acc_std_relax', 'sc_forces']:
            raise NotImplementedError('Specified workflow currently not implemented.')
        else:
            self._calculation_workflow = ''.join(calculation_workflow.title().split('_'))

    @property
    def custom_calculation_settings(self):
        return self._custom_calculation_settings

    @custom_calculation_settings.setter
    def custom_calculation_settings(self, custom_calculation_settings):
        if not custom_calculation_settings:
            self._custom_calculation_settings = {}
        elif not os.path.isfile(custom_calculation_settings):
            error_message = 'Specified custom settings file {} not found'.format(custom_calculation_settings)
            raise KelpieGrazerError(error_message)
        else:
            with open(custom_calculation_settings, 'r') as fr:
                self._custom_calculation_settings = json.load(fr)

    @property
    def mpi_call_file(self):
        return self._mpi_call_file

    @mpi_call_file.setter
    def mpi_call_file(self, mpi_call_file):
        if not mpi_call_file:
            mpi_call_file = os.path.join(self.run_location, 'mpi_call.txt')
        if not os.path.exists(mpi_call_file):
            error_message = 'MPI call file {} not found'.format(mpi_call_file)
            raise KelpieGrazerError(error_message)
        else:
            self._mpi_call_file = os.path.abspath(mpi_call_file)

    @property
    def mpi_call(self):
        with open(self.mpi_call_file, 'r') as fr:
            return fr.read()

    def graze(self):
        if not os.path.isdir(self.run_location):
            os.makedirs(self.run_location)

        workflow = getattr(calculation_workflows, '{}Workflow'.format(self.calculation_workflow))(
            initial_structure=self.initial_structure,
            run_location=self.run_location,
            custom_calculation_settings=self.custom_calculation_settings,
            mpi_call=self.mpi_call,
            **self.kwargs)
        workflow.perform_workflow()



import os
from kelpie.structure import Atom
from kelpie.structure import Structure
from kelpie.data import STD_ATOMIC_WEIGHTS


class KelpieIOError(Exception):
    """Base class for I/O related error handling."""
    pass


def read_poscar(poscar_file='POSCAR'):
    """
    :param poscar_file: Location of the VASP POSCAR (version 5) file
                        NOTE: Names of all species (line 6) need to begin with the symbol of a real element.
    :return: `kelpie.structure.Structure` object
    """
    if not os.path.isfile(poscar_file):
        error_message = 'Specified POSCAR file {} not found'.format(poscar_file)
        raise FileNotFoundError(error_message)

    with open(poscar_file, 'r') as fr:
        poscar_lines = [line.strip() for line in fr.readlines()]

    if not _consistent_number_of_atoms(poscar_lines):
        error_message = 'Mismatch between the number of atoms (Line 7) and the number of atomic coordinates (Lines 9-)'
        raise KelpieIOError(error_message)

    poscar_blocks = ['system_title',
                     'scaling_constant',
                     'lattice_vectors',
                     'list_of_species',
                     'list_of_number_of_atoms',
                     'repeating_list_of_species',
                     'coordinate_system',
                     'list_of_atomic_coordinates'
                     ]

    poscar_as_dict = {}
    for block in poscar_blocks:
        poscar_as_dict[block] = globals()['_{}'.format(block)](poscar_lines)

    s = Structure()
    s.comment = poscar_as_dict['system_title']
    s.scaling_constant = poscar_as_dict['scaling_constant']
    s.lattice_vectors = poscar_as_dict['lattice_vectors']
    s.coordinate_system = poscar_as_dict['coordinate_system']
    for ac, sp in zip(poscar_as_dict['list_of_atomic_coordinates'], poscar_as_dict['repeating_list_of_species']):
        atom = Atom(coordinates=ac, species=sp)
        s.add_atom(atom)

    return s


def _consistent_number_of_atoms(poscar_lines):
    """Is the total number of atoms (Line 7) consistent with the number of atomic coordinates specified?

    :return: Boolean """
    return len(_repeating_list_of_species(poscar_lines)) == len(_list_of_atomic_coordinates(poscar_lines))


def _system_title(poscar_lines):
    """:return: String with the system title
    """
    return poscar_lines[0]


def _scaling_constant(poscar_lines):
    """ Parse the scaling constant for the structure (line 2).

    :return: Float with the scaling constant
    :raise KelpieIOError: if the scaling constant cannot be converted to float
    """
    try:
        scaling_constant = float(poscar_lines[1])
    except ValueError:
        error_message = 'Scaling factor (Line 2 in POSCAR) should be float'
        raise KelpieIOError(error_message)
    else:
        return scaling_constant


def _lattice_vectors(poscar_lines):
    """Parse the lattice vectors of the structure (lines 3-5).

    :return: 3x3-shaped List of Float with the lattice vectors [[a11, a12, a13], [a21, a22, a23], ...]
    :raise KelpieIOError: if any lattice vector component cannot be converted to float
    """
    lattice_vectors = []
    try:
        for line in poscar_lines[2:5]:
            lattice_vectors.append([float(a) for a in line.split()])
    except ValueError:
        error_message = 'Lattice vector components (Lines 3-5) should be floating point'
        raise KelpieIOError(error_message)
    else:
        return lattice_vectors


def _list_of_species(poscar_lines):
    """Parse the species in the structure (line 6).

    :return: List of String with the species symbols
    :raise KelpieIOError: if any of the species names contains only integers
    """
    species_list = poscar_lines[5].split()
    for species in species_list:
        if not any([species.startswith(e) for e in STD_ATOMIC_WEIGHTS]):
            error_message = 'All species names (Line 6) must begin with the symbol of a real element.'
            error_message += '\nNOTE: Only VASP 5 POSCAR format is supported'
            raise KelpieIOError(error_message)
    return species_list


def _list_of_number_of_atoms(poscar_lines):
    """Parse the number of atoms of each species in the structure (line 7).

    :return: List of Int with the number of atoms of each species
    :raise KelpieIOError: if any of the number of atoms cannot be converted to int
    """
    try:
        list_of_number_of_atoms = [int(n) for n in poscar_lines[6].split()]
    except ValueError:
        error_message = 'Number of atoms of each species (Line 7) should be integers'
        raise KelpieIOError(error_message)
    else:
        return list_of_number_of_atoms


def _repeating_list_of_species(poscar_lines):
    repeating_list_of_species = []
    for s, n in zip(_list_of_species(poscar_lines), _list_of_number_of_atoms(poscar_lines)):
        repeating_list_of_species += [s]*n
    return repeating_list_of_species


def _coordinate_system(poscar_lines):
    """
    Are the atomic positions in direct/fractional or cartesian coordinates? (line 8)

    :return: "Direct" or "Cartesian"
    :raise NotImplementedError: for Selective Dynamics
    :raise KelpieIOError: if coordinate system is not VASP-recognizable
    """
    coordinate_system = poscar_lines[7]
    first_char = coordinate_system.lower()[0]  # VASP only recognizes the first character
    if first_char == 'd':
        return 'Direct'
    elif first_char in ['c', 'k']:
        return 'Cartesian'
    elif first_char == 's':
        error_message = 'Selective dynamics I/O handling currently not implemented'
        raise NotImplementedError(error_message)
    else:
        error_message = 'Coordinate system (Line 8) can only be direct or cartesian'
        raise KelpieIOError(error_message)


def _list_of_atomic_coordinates(poscar_lines):
    """
    Parse all the atomic coordinates (line 9-(9+number of atoms)).

    :return: Nx3-shaped List of Float with the atomic coordinates [[c11, c12, c13], [c21, c22, c23], ...]
    :raise KelpieIOError: if any atomic coordinate component cannot be converted to float
    """
    atomic_coordinates = []
    for i, line in enumerate(poscar_lines[8:]):
        if line[0] == '#':
            break
        try:
            coord = [float(c) for c in line.split()[:3]]
        except ValueError:
            error_message = 'Check the atomic coordinates block (Line {}) for non-float values'.format(9+i)
            raise KelpieIOError(error_message)
        else:
            atomic_coordinates.append(coord)
        if i == sum(_list_of_number_of_atoms(poscar_lines))-1:
            break
    return atomic_coordinates



import os
import math
import six
import kelpie
from kelpie import io
from kelpie.structure import Structure
from kelpie.vasp_settings.incar import DEFAULT_VASP_INCAR_SETTINGS, VASP_INCAR_TAGS, HUBBARD_U_VALUES
from kelpie.vasp_settings.potcar import VASP_RECO_POTCARS
from kelpie.data import STD_ATOMIC_WEIGHTS, VALENCE_ELECTRONS


class VaspInputError(Exception):
    """Base class for errors in VASP input files."""
    pass


class VaspInputGenerator(object):
    """Base class to generate VASP input files for a given POSCAR file."""

    def __init__(self,
                 structure=None,
                 calculation_settings=None,
                 write_location=None,
                 **kwargs):
        """
        :param structure: `kelpie.structure.Structure` object or String with the location of a valid VASP5 POSCAR file
        :param calculation_settings: Dictionary of VASP INCAR tags, values, and POTCAR choices.
                                     Defaults to pre-defined full relaxation calculation.
        :param write_location: String with the location where the VASP input files should be written.
                               Defaults to current working directory
        :param kwargs: Dictionary of other miscellaneous parameters, if any.
        """

        #: `kelpie.structure.VaspStructure` object containing the VASP POSCAR data
        self._structure = None
        self.structure = structure

        #: VASP INCAR tags, values and POTCAR choices for generating input files.
        self._calculation_settings = None
        self.calculation_settings = calculation_settings

        #: where the VASP input files should be written
        self._write_location = None
        self.write_location = write_location

        #: other miscellaneous keyword arguments
        self.kwargs = kwargs

    @property
    def structure(self):
        return self._structure

    @structure.setter
    def structure(self, structure):
        if isinstance(structure, Structure):
            structure.check_structure_is_complete()
            self._structure = structure
        elif isinstance(structure, six.string_types):
            if os.path.isfile(structure):
                structure = io.read_poscar(poscar_file=structure)
                self._structure = structure
            else:
                error_message = 'POSCAR file {} not found'.format(structure)
                raise VaspInputError(error_message)
        else:
            error_message = '`structure` must be `structure.Structure` object or path to a valid POSCAR'
            raise VaspInputError(error_message)

    @property
    def calculation_settings(self):
        return self._calculation_settings

    @calculation_settings.setter
    def calculation_settings(self, calculation_settings):
        if not calculation_settings:
            self._calculation_settings = DEFAULT_VASP_INCAR_SETTINGS['relaxation']
        else:
            self._calculation_settings = calculation_settings

        # if ENCUT is not set in calculation settings, set it manually
        if self._calculation_settings.get('encut') is None:
            self.set_calculation_encut()

        # Scale EDIFF with the number of atoms in the unit cell
        if self._calculation_settings.get('scale_ediff_per_atom'):
            self.scale_ediff_per_atom()

        # if the structure contains elements with unfilled d or f shells, set ISPIN = 2, and
        # appropriate MAGMOM tag
        if self._calculation_settings.get('magnetism', None) is not None:
            self.set_magnetism()

        # if the structure contains O and one of the TMs for which Hubbard U value to use is known,
        # set the appropriate VASP INCAR TAGS
        hubbard_values_scheme = self._calculation_settings.get('hubbards', None)
        if hubbard_values_scheme is not None:
            self.set_hubbards(scheme=hubbard_values_scheme)

    @property
    def write_location(self):
        return self._write_location

    @write_location.setter
    def write_location(self, write_location):
        if not write_location:
            self._write_location = os.path.abspath(os.getcwd())
        else:
            self._write_location = write_location

    def vasp_tag_value_formatter(self, value):
        if isinstance(value, list):
            return ' '.join(map(self.vasp_tag_value_formatter, value))
        elif isinstance(value, six.string_types):
            return value.upper()
        elif isinstance(value, bool):
            return '.{}.'.format(str(value).upper())
        elif isinstance(value, float):
            return '{:.2E}'.format(value)
        else:
            return str(value)

    def format_vasp_tag(self, tag, value):
        """Format INCAR tags and corresponding values to be printed in the INCAR.

        :param tag: VASP INCAR tag
        :type tag: str
        :param value: value corresponding to the INCAR tag
        :type value: str or bool or float or list(str or bool or float)
        :return: formatted tag, value
        :rtype: str
        """
        return '{:14s} = {}'.format(tag.upper(), self.vasp_tag_value_formatter(value))

    @staticmethod
    def _concatenate_potcars(list_of_potcar_files):
        """Concatenate a list of POTCAR files into a single POTCAR.

        :param list_of_potcar_files: relative/absolute path of a list of VASP POTCAR files.
        :type list_of_potcar_files: list(str)
        :return: contents of the concatenated VASP POTCAR
        :rtype: str

        """
        concatenated_potcar = ''
        for potcar_file in list_of_potcar_files:
            with open(potcar_file, 'r') as fr:
                concatenated_potcar += fr.read()
        return concatenated_potcar

    @property
    def POTCAR(self):
        """VASP POTCAR file for `self.vasp_structure` and `self.calculation settings`.

        :return: contents of the VASP POTCAR file.
        :rtype: str
        """
        return self._get_vasp_potcar()

    def _get_vasp_potcar(self):
        """Construct the VASP POTCAR for `self.POSCAR`.

        If `self.calculation_settings['potcar_settings']['element_potcars'] dictionary:
        (a) is empty: defaults are used from `kelpie.vasp_settings.potcar.VASP_RECO_POTCARS`.
        (b) has actual location of POTCARs as values to element keys, they are used directly.
        (c) has VASP POTCAR tag (e.g. "Mn_pv") as values to element keys, corresponding POTCARs are used from the
            local POTCAR library.

        :return: VASP POTCAR file
        :rtype: str
        :raise VaspInputError: if POTCAR file for any element is not found.
        """
        potcar_settings = self.calculation_settings['potcar_settings']
        version = potcar_settings['version']
        xc = potcar_settings['xc']
        element_potcars = potcar_settings.get('element_potcars', {})

        list_of_potcars = []
        for species in self.structure.list_of_species:
            element_potcar = element_potcars.get(species)

            # if POTCAR choices are not specified in the calculation settings, use defaults.
            if element_potcar is None:
                potcar_dir = VASP_RECO_POTCARS[version][xc]['path']
                reco_pots = VASP_RECO_POTCARS[version][xc]['reco']
                try:
                    potcar_file = os.path.join(potcar_dir, reco_pots[species], 'POTCAR')
                except KeyError:
                    error_message = 'POTCAR for {} not found'.format(species)
                    raise VaspInputError(error_message)
                else:
                    list_of_potcars.append(potcar_file)

            # if actual location of the POTCARs are specified in the calculation settings, use them.
            elif os.path.isfile(element_potcar):
                list_of_potcars.append(element_potcar)

            # if VASP POTCAR keys (e.g. "Li_sv", "Mn_pv") are specified, read them from the local POTCAR library.
            else:
                potcar_dir = VASP_RECO_POTCARS[version][xc]['path']
                potcar_file = os.path.join(potcar_dir, element_potcar, 'POTCAR')
                if os.path.isfile(potcar_file):
                    list_of_potcars.append(potcar_file)
                else:
                    error_message = 'POTCAR file {} for {} not found'.format(potcar_file, species)
                    raise VaspInputError(error_message)
        return self._concatenate_potcars(list_of_potcars)

    @staticmethod
    def get_highest_enmax(vasp_potcar):
        """Make a list of all ENMAX values in the POTCAR, return the highest among them.

        :param vasp_potcar: VASP POTCAR file
        :type vasp_potcar: str (with linebreaks)
        :return: highest ENMAX value from the POTCAR
        :rtype: float
        """
        list_of_enmax = []
        for line in vasp_potcar.split('\n'):
            if 'ENMAX' in line:
                list_of_enmax.append(float(line.strip().split()[2].strip(';')))
        return max(list_of_enmax)

    @staticmethod
    def roundup_encut(encut):
        """Roundup the encut value to the nearest ten.
        :param encut: ENCUT for VASP INCAR
        :type encut: float
        :return: encut rounded up to the nearest ten
        :rtype: int
        """
        return int(math.ceil(encut/10.))*10

    def set_calculation_encut(self):
        """Update `self.calculation_settings['encut']` = encut_scaling_factor*(highest ENMAX from `self.POTCAR`).

        If `encut_scaling_factor` is not specified in calculation settings, 1.5 is used as default.
        If `maximum_encut` is not specified in calculation settings, 600 (in eV) is used as default.
        """
        encut_scaling_factor = self.calculation_settings.get('encut_scaling_factor', 1.5)
        maximum_encut = self.calculation_settings.get('maximum_encut', 600)
        encut = self.roundup_encut(encut_scaling_factor*self.get_highest_enmax(self.POTCAR))
        if encut > maximum_encut:
            encut = maximum_encut
        self._calculation_settings.update({'encut': encut})

    def scale_ediff_per_atom(self):
        """Scale EDIFF with number of atoms in the unit cell. EDIFF' = EDIFF*number of atoms in the cell."""
        ediff = self._calculation_settings.get('ediff', DEFAULT_VASP_INCAR_SETTINGS['relaxation']['ediff'])
        self._calculation_settings.update({'ediff': ediff*self.structure.natoms, 'scale_ediff_per_atom': False})

    def set_magnetism(self):
        if self._calculation_settings['magnetism'] != 'ferro':
            error_message = 'Only ferromagnetic initial configuration implemented'
            raise NotImplementedError(error_message)
        for atom in self.structure.atoms:
            element = sorted([e for e in STD_ATOMIC_WEIGHTS if atom.species.startswith(e)])[-1]
            if 0 < VALENCE_ELECTRONS[element]['d_elec'] < 10:
                atom.magmom = 5.0
            if 0 < VALENCE_ELECTRONS[element]['f_elec'] < 14:
                atom.magmom = 7.0
        self.set_ispin()
        self.set_magmom_tag()

    def set_ispin(self):
        if any([a.magmom for a in self.structure.atoms]):
            self._calculation_settings.update({'ispin': 2})
        else:
            self._calculation_settings.update({'ispin': 1})

    def set_magmom_tag(self):
        if self._calculation_settings.get('ispin') == 1:
            return
        else:
            self._calculation_settings.update({'magmom': self.structure.MAGMOM})

    def get_hubbards(self, scheme='wang'):
        if scheme not in HUBBARD_U_VALUES:
            error_message = 'Cannot find the specified scheme "{}" for Hubbard U values to use'.format(scheme)
            raise VaspInputError(error_message)
        if 'O' not in self.structure.list_of_species:
            hubbards = {'ldau': False}
            return hubbards
        ldaul = []
        ldauu = []
        ldauj = []
        for species in self.structure.list_of_species:
            element = sorted([e for e in STD_ATOMIC_WEIGHTS if species.startswith(e)])[-1]
            if element in HUBBARD_U_VALUES[scheme]:
                ldaul.append(2)
                ldauu.append(HUBBARD_U_VALUES[scheme][element])
            else:
                ldaul.append(-1)
                ldauu.append(0.)
            ldauj.append(0.)
        # if no species has a Hubbard U value to use, switch off LDAU
        if all([v == -1 for v in ldaul]):
            hubbards = {'ldau': False}
            return hubbards
        # otherwise, generate the VASP INCAR TAGS and return as a Dictionary
        ldaul = ' '.join(['{}'.format(v) for v in ldaul])
        ldauu = ' '.join(['{:.2f}'.format(v) for v in ldauu])
        ldauj = ' '.join(['{:.2f}'.format(v) for v in ldauj])
        hubbards = {
            'ldau': True,
            'ldautype': 2,
            'ldaul': ldaul,
            'ldauu': ldauu,
            'ldauj': ldauj,
            'ldaprint': 1,
            'lmaxmix': 4
        }
        return hubbards

    def set_hubbards(self, scheme='wang'):
        self._calculation_settings.update(self.get_hubbards(scheme=scheme))

    @property
    def INCAR(self):
        """VASP INCAR for the `self.vasp_structure` and `self.calculation_settings`.

        :return: contents of the VASP INCAR file.
        :rtype: str
        """
        return self._get_vasp_incar()

    def _get_vasp_incar(self):
        """Construct the VASP INCAR file for `self.POSCAR` using `self.calculation_settings`

        :return: VASP INCAR file
        :rtype: str
        """
        vasp_incar = ''
        for tag_block in VASP_INCAR_TAGS['tags']:
            vasp_incar += '### {} ###\n'.format(tag_block)
            for tag in VASP_INCAR_TAGS[tag_block]:
                value = self.calculation_settings.get(tag)
                if value is None:
                    continue
                vasp_incar += self.format_vasp_tag(tag, value) + '\n'
            vasp_incar += '\n'
        vasp_incar += '# [autogenerated by kelpie v{}] #'.format(kelpie.__version__)
        return vasp_incar

    @staticmethod
    def _write_vasp_input_file(file_contents, file_location, overwrite=True):
        """Write the `file_contents` as is into `file_location`; confirm overwrite if file already exists.

        :param file_contents: contents to be written into file
        :type file_contents: str
        :param file_location: absolute or relative path to the file
        :type file_location: str
        :param overwrite: if `file_location` already exists, should it be overwritten?
        """
        if os.path.isfile(file_location) and not overwrite:
            return
        with open(file_location, 'w') as fw:
            fw.write(file_contents)
        return

    def write_vasp_input_files(self, overwrite=True):
        """Write VASP POSCAR, POTCAR, and INCAR files into the folder specified by `self.write_location`.

        :param overwrite: Should files be overwritten if they already exist? Defaults to True."""
        # write the POSCAR file
        poscar_file = os.path.join(self.write_location, 'POSCAR')
        self._write_vasp_input_file(self.structure.POSCAR, poscar_file, overwrite=overwrite)

        # write the POTCAR file
        potcar_file = os.path.join(self.write_location, 'POTCAR')
        self._write_vasp_input_file(self.POTCAR, potcar_file, overwrite=overwrite)

        # write the INCAR file
        incar_file = os.path.join(self.write_location, 'INCAR')
        self._write_vasp_input_file(self.INCAR, incar_file, overwrite=overwrite)




import os
import gzip
import numpy as np
import datetime
from lxml import etree


class VasprunXMLParserError(Exception):
    """Base class to handle errors related to parsing the vasprun.xml file."""
    pass


class VasprunXMLParser(object):
    """Base class to parse relevant output from a vasprun.xml file."""

    def __init__(self, vasprunxml_file=None):
        """
        :param vasprunxml_file: name of the vasprun.xml file (default: None)
        :type vasprunxml_file: str or None
        """
        self._vasprunxml_file = None
        self._xmlroot = None

        self.vasprunxml_file = vasprunxml_file

    @property
    def vasprunxml_file(self):
        return self._vasprunxml_file

    @vasprunxml_file.setter
    def vasprunxml_file(self, vasprunxml_file):
        if vasprunxml_file is None:
            return
        if not os.path.isfile(vasprunxml_file):
            error_message = 'Cannot find the specified vasprun.xml file: {}'.format(vasprunxml_file)
            raise VasprunXMLParserError(error_message)
        self._vasprunxml_file = os.path.abspath(vasprunxml_file)
        self._xmlroot = self._get_vasprunxml_root()

    def _get_vasprunxml_root(self):
        """Read contents from a vasprun.xml or vasprun.xml.gz file, convert it into
        etree.ElementTree and get the root element with tag 'modeling'

        :raises: VasprunXMLParserError if the root element is not 'modeling'
        :return: root element of vasprun.xml
        :rtype: etree._Element
        """
        if self.vasprunxml_file is None:
            return
        parser = etree.XMLParser(remove_blank_text=True)
        try:
            tree = etree.parse(self.vasprunxml_file, parser=parser)
        except etree.XMLSyntaxError:
            return
        xmlroot = tree.getroot()
        if xmlroot.tag != 'modeling':
            error_message = 'Root element of vasprun.xml "modeling" not found'
            raise VasprunXMLParserError(error_message)
        return xmlroot

    @property
    def xmlroot(self):
        return self._xmlroot

    def read_composition_information(self):
        """Read the list of elemental species in the unit cell, and number of atoms, atomic mass, number of valence
        electrons, VASP pseudopotential title tag for each species.

        :return: unit cell composition information.
                 - {element1: {'natoms': n1, 'atomic_mass': m1, 'valence': v1, 'pseudopotential': p1}, element2: ...}
        :rtype: dict(str, dict(str, int or float or str))
        """
        composition_info = {}
        if self.xmlroot is not None:
            atomtypes_array = self.xmlroot.findall('./atominfo/array')
            for array in atomtypes_array:
                if array.attrib['name'] != 'atomtypes':
                    continue
                for species in array.findall('./set/rc'):
                    natoms, elem, mass, valence, psp = [c.text.strip() for c in species.findall('c')]
                    composition_info.update({elem: {'natoms': int(natoms),
                                                    'atomic_mass': float(mass),
                                                    'valence': float(valence),
                                                    'pseudopotential': psp
                                                    }
                                            })
        return composition_info

    def read_list_of_atoms(self):
        """Read the list of atoms in the unit cell.

        :return: list of atoms ['atom1', 'atom1', 'atom2', 'atom2', 'atom2', ...]
        :rtype: list
        """
        atomslist = []
        if self.xmlroot is not None:
            atoms_array = self.xmlroot.findall('./atominfo/array')
            for array in atoms_array:
                if array.attrib['name'] != 'atoms':
                    continue
                for species in array.findall('./set/rc'):
                    atom_symbol, atomtype = [c.text.strip() for c in species.findall('c')]
                    atomslist.append(atom_symbol)
        return atomslist

    def read_number_of_ionic_steps(self):
        """Read number of ionic steps in the VASP run.

        :return: number of ionic steps
        :rtype: int
        """
        if self.xmlroot is not None:
            return len(self.xmlroot.findall('./calculation'))

    def read_scf_energies(self):
        """Read all the the energies in every ionic step.

        :return: {ionic_step_1: [e1, e2, e3, ...], ionic_step_2: [e1, e2, ...], ionic_step_3: ...}
        :rtype: dict(int, list(float))
        """
        scf_energies = {}
        if self.xmlroot is not None:
            ionic_steps = self.xmlroot.findall('./calculation')
            for n_ionic_step, ionic_step in enumerate(ionic_steps):
                scsteps = ionic_step.findall('scstep')
                scstep_energies = []
                for scstep in scsteps:
                    for energy in scstep.findall('./energy/i'):
                        if energy.attrib['name'] == 'e_fr_energy':
                            scstep_energies.append(float(energy.text.strip()))
                scf_energies[n_ionic_step] = scstep_energies
        return scf_energies

    def read_entropies(self):
        """Read entropy at the end of each ionic step.

        :return: {ionic_step_1: entropy_1, ionic_step_2: entropy_2, ionic_step_3: ...}
        :rtype: dict(int, float)
        """
        entropy_dict = {}
        if self.xmlroot is not None:
            ionic_steps = self.xmlroot.findall('./calculation')
            for n_ionic_step, ionic_step in enumerate(ionic_steps):
                entropy = None
                final_scstep = ionic_step.findall('scstep')[-1]
                for final_energy_block in final_scstep.findall('energy'):
                    for energy in final_energy_block.findall('i'):
                        if energy.attrib['name'] == 'eentropy':
                            entropy = float(energy.text.strip())
                entropy_dict[n_ionic_step] = entropy
        return entropy_dict

    def read_free_energies(self):
        """Read free energy at the end of each ionic step.

        :return: {ionic_step_1: free_energy_1, ionic_step_2: free_energy_2, ionic_step_3: ...}
        :rtype: dict(int, float)
        """
        free_energy_dict = {}
        if self.xmlroot is not None:
            ionic_steps = self.xmlroot.findall('./calculation')
            for n_ionic_step, ionic_step in enumerate(ionic_steps):
                free_energy = None
                final_scstep = ionic_step.findall('scstep')[-1]
                for final_energy_block in final_scstep.findall('energy'):
                    for energy in final_energy_block.findall('i'):
                        if energy.attrib['name'] == 'e_fr_energy':
                            free_energy = float(energy.text.strip())
                free_energy_dict[n_ionic_step] = free_energy
        return free_energy_dict

    def read_forces(self):
        """Read forces on all atoms in the unit cell at the end of each ionic step.

        :return: {ionic_step_1: [[fx_1, fy_1, fz_1], [fx_2, fy_2, fz_2], ...], ionic_step_2: ...}
        :rtype: dict(int, numpy.array)
                - numpy.array of shape (N_atoms, 3)
        """
        forces_dict = {}
        if self.xmlroot is not None:
            ionic_steps = self.xmlroot.findall('./calculation')
            for n_ionic_step, ionic_step in enumerate(ionic_steps):
                varrays = ionic_step.findall('varray')
                forces = []
                for varray in varrays:
                    if varray.attrib['name'] != 'forces':
                        continue
                    for force_on_atom in varray.findall('v'):
                        forces.append([float(e) for e in force_on_atom.text.split()])
                forces_dict[n_ionic_step] = np.array(forces)
        return forces_dict

    def read_stress_tensors(self):
        """Read stress (in kbar) on the unit cell at the end of each ionic step.

        :return: {ionic_step_1: [[Sxx, Sxy, Sxz], [Syx, Syy, Syz], [Szx, Szy, Szz]], ionic_step_2: ...}
        :rtype: dict(int, numpy.array)
                - numpy.array of shape (3, 3)
        """
        stress_tensor_dict = {}
        if self.xmlroot is not None:
            ionic_steps = self.xmlroot.findall('./calculation')
            for n_ionic_step, ionic_step in enumerate(ionic_steps):
                varrays = ionic_step.findall('varray')
                stress_tensor = []
                for varray in varrays:
                    if varray.attrib['name'] != 'stress':
                        continue
                    for stress_component in varray.findall('v'):
                        stress_tensor.append([float(e) for e in stress_component.text.split()])
                stress_tensor_dict[n_ionic_step] = np.array(stress_tensor)
        return stress_tensor_dict

    def read_lattice_vectors(self):
        """Read lattice vectors (in Angstrom) of the unit cell at the end of each ionic step.

        :return: {ionic_step_1: [[a11, a12, a13], [a21, a22, a23], [a31, a32, a33]], ionic_step_2: ...}
        :rtype: dict(key, numpy.array)
                - numpy.array of shape (3, 3)
        """
        lattice_vectors_dict = {}
        if self.xmlroot is not None:
            ionic_steps = self.xmlroot.findall('./calculation')
            for n_ionic_step, ionic_step in enumerate(ionic_steps):
                varrays = ionic_step.findall('./structure/crystal/varray')
                lattice_vectors = []
                for varray in varrays:
                    if varray.attrib['name'] != 'basis':
                        continue
                    for lattice_vector in varray.findall('v'):
                        lattice_vectors.append([float(e) for e in lattice_vector.text.split()])
                lattice_vectors_dict[n_ionic_step] = np.array(lattice_vectors)
        return lattice_vectors_dict

    def read_atomic_coordinates(self):
        """Read positions of all the atoms at the end of each ionic step.

        :return: {ionic_step_1: [[species_1, [x1, y1, z1]], [species_2, [x2, y2, z2]], ...], ionic_step_2: ...}
        :rtype: dict(int, list(list(str, numpy.array)))
        """
        atomic_coordinates = {}
        if self.xmlroot is not None:
            ionic_steps = self.xmlroot.findall('calculation')
            for n_ionic_step, ionic_step in enumerate(ionic_steps):
                atomic_coordinates[n_ionic_step] = []
                varray = ionic_step.find('./structure/varray')
                if varray.attrib['name'] != 'positions':
                    continue
                for coordinates in varray.findall('v'):
                    atomic_coordinates[n_ionic_step].append([float(e) for e in coordinates.text.split()])
        return atomic_coordinates

    def read_cell_volumes(self):
        """Read the volume (in cubic Angstrom) of the unit cell at the end of each ionic step.

        :return: {ionic_step_1: float, ionic_step_2: float}
        :rtype: dict(int, float)
        """
        volume_dict = {}
        if self.xmlroot is not None:
            ionic_steps = self.xmlroot.findall('calculation')
            for n_ionic_step, ionic_step in enumerate(ionic_steps):
                volume = float(ionic_step.find('./structure/crystal/i').text.strip())
                volume_dict[n_ionic_step] = volume
        return volume_dict

    def read_kpoint_mesh(self):
        """Read the k-point mesh (k_x, k_y, k_z) used in the calculation.

        :return: [k_x, k_y, k_z]
        :rtype: list(int)
        """
        if self.xmlroot is not None:
            kpoints_block = self.xmlroot.find('kpoints')
            for v in kpoints_block.findall('./generation/v'):
                if v.attrib['name'] == 'divisions':
                    kmesh = [int(k) for k in v.text.split()]
                    return kmesh

    def read_irreducible_kpoints(self):
        """Read all the irreducible k-points used in the calculation.

        :return: [[kx_1, ky_1, kz_1], [kx_2, ky_2, kz_2], ...]
        :rtype: list(list(float))
        """
        kpoints = []
        if self.xmlroot is not None:
            kpoints_block = self.xmlroot.find('kpoints')
            for kpointlist in kpoints_block.findall('varray'):
                if kpointlist.attrib['name'] == 'kpointlist':
                    for v in kpointlist.findall('v'):
                        kpoints.append([float(k) for k in v.text.split()])
        return kpoints

    def read_total_density_of_states(self):
        """
        :return: Total density of states data {'spin_1': [[energy1, dos1, intdos_1], ...], 'spin_2': ...}
        :rtype: dict(str, list(list(float)))
        """
        total_dos_data = {}
        if self.xmlroot is not None:
            dos_block = self.xmlroot.find('./calculation/dos')
            for spin in dos_block.findall('./total/array/set/set'):
                spin_data = []
                for gridpoint in spin.findall('r'):
                    spin_data.append([float(e) for e in gridpoint.text.strip().split()])
                total_dos_data[spin.attrib['comment'].replace(' ', '_')] = spin_data
        return total_dos_data

    def read_fermi_energy(self):
        """
        :return: the Fermi energy in eV
        :rtype: float or None
        """
        try:
            fermi_energy = float(self.xmlroot.find('./calculation/dos/i').text.strip())
        except (AttributeError, TypeError) as error:
            fermi_energy = None
        return fermi_energy

    def read_band_occupations(self):
        """Read occupation of every band at every k-point for each spin channel.

        :return: {'spin_1': {kpoint_1: {'band_energy': [band1, ...], 'occupation': [occ1, ...]}, 'kpoint_2': ...}}
        :rtype: dict(str, dict(int, dict(str, list(float))))
        """
        occupations_dict = {}
        if self.xmlroot is not None:
            final_ionic_step = self.xmlroot.findall('calculation')[-1]
            eigenvalues = final_ionic_step.find('eigenvalues')
            if eigenvalues is not None:
                for spin_set in eigenvalues.findall('./array/set/set'):
                    spin = spin_set.attrib['comment'].replace(' ', '_')
                    occupations_dict[spin] = {}
                    for kpoint_set in spin_set.findall('./set'):
                        kpoint = int(kpoint_set.attrib['comment'].split()[-1])
                        occupations_dict[spin][kpoint] = {'band_energy': [], 'occupation': []}
                        for band in kpoint_set.findall('./r'):
                            be, occ = [float(b) for b in band.text.strip().split()]
                            occupations_dict[spin][kpoint]['band_energy'].append(be)
                            occupations_dict[spin][kpoint]['occupation'].append(occ)
        return occupations_dict

    def read_run_timestamp(self):
        """Read the time and date when the calulation was run.

        :return: year, month, day, hour, minute, second when the calculation was run.
        :rtype: `datetime.datetime` object
        """
        if self.xmlroot is not None:
            date_and_time = self.xmlroot.findall('./generator/i')
            year = month = day = hour = minute = second = 0
            for field in date_and_time:
                if field.attrib['name'] == 'date':
                    year, month, day = [int(f) for f in field.text.strip().split()]
                if field.attrib['name'] == 'time':
                    hour, minute, second = [int(f) for f in field.text.strip().split(':')]
            return datetime.datetime(year=year, month=month, day=day, hour=hour, minute=minute, second=second)

    def read_scf_looptimes(self):
        """Read total time taken for each SCF loop during the run.

        :return: {ionic_step_1: [t1, t2, t3, ...], ionic_step_2: [t1, t2, ...], ...}
        :rtype: dict(int, list(float))
        """
        scf_looptimes = {}
        if self.xmlroot is not None:
            ionic_steps = self.xmlroot.findall('./calculation')
            for n_ionic_step, ionic_step in enumerate(ionic_steps):
                scsteps = ionic_step.findall('scstep')
                scstep_times = []
                for scstep in scsteps:
                    for time in scstep.findall('time'):
                        if time.attrib['name'] == 'total':
                            scstep_times.append(float(time.text.strip().split()[-1]))
                scf_looptimes[n_ionic_step] = scstep_times
        return scf_looptimes


class VaspOutcarParserError(Exception):
    """Base class to handle errors related to parsing the VASP OUTCAR file."""


class VaspOutcarParser(object):
    """Base class to parse relevant output from a VASP OUTCAR file. Currently only for data not present in the
    vasprun.xml file. For example, atom-projected charge and magnetic moment information.
    """

    def __init__(self, vasp_outcar_file=None):
        """
        :param vasp_outcar_file: name of the VASP OUTCAR file (default: None)
        :type vasp_outcar_file: str or None
        """
        self._vasp_outcar_file = None
        self._outcar = None

        self.vasp_outcar_file = vasp_outcar_file
        self._n_ions = self._get_n_ions()

    @property
    def vasp_outcar_file(self):
        return self._vasp_outcar_file

    @vasp_outcar_file.setter
    def vasp_outcar_file(self, vasp_outcar_file):
        if vasp_outcar_file is None:
            return
        if not os.path.isfile(vasp_outcar_file):
            error_message = 'Cannot find the specified VASP OUTCAR file: {}'.format(vasp_outcar_file)
            raise VaspOutcarParserError(error_message)
        self._vasp_outcar_file = os.path.abspath(vasp_outcar_file)
        self._outcar = self._get_outcar_contents()

    def _get_outcar_contents(self):
        """Return the contents of a given VASP OUTCAR file."""
        if self.vasp_outcar_file is None:
            return
        if os.path.splitext(self.vasp_outcar_file)[-1] == '.gz':
            try:
                with gzip.open(self.vasp_outcar_file, 'rb') as fr:
                    return fr.read().decode('utf-8')
            except OSError:
                pass
        with open(self.vasp_outcar_file, 'r') as fr:
            return fr.read()

    @property
    def outcar(self):
        return self._outcar

    def _get_n_ions(self):
        if self.outcar is None:
            return
        outcar_lines = self.outcar.splitlines()
        for line in outcar_lines:
            if 'NIONS' in line:
                n_ions = int(line.strip().split()[-1])
                return n_ions

    @property
    def n_ions(self):
        return self._n_ions

    def read_orb_projected_charge(self):
        """Read the orbital-projected charge block (example below) in a VASP OUTCAR file, if present.

         total charge

        # of ion       s       p       d       tot
        ------------------------------------------
            1        0.476   0.546   7.466   8.488
            2        0.476   0.545   7.465   8.486
            3        2.192   6.110   1.929  10.231
            4        2.192   6.110   1.935  10.237
            ...
           12        0.470   0.560   6.454   7.485
        --------------------------------------------------
        tot         14.948  32.175  35.925  83.048

        :return: {atom1: {'s': charge_s1, 'p': charge_p1, ...}, 2: {...}, ...}
        :rtype: dict(int, dict(str, float))
        """
        if self.outcar is None:
            return
        outcar_lines = self.outcar.splitlines()
        start_index = None
        for r_index, line in enumerate(outcar_lines[::-1]):
            if 'total charge' in line:
                if '# of ion' in outcar_lines[len(outcar_lines) - r_index + 1]:
                    start_index = len(outcar_lines) - r_index - 1
                    break
        if start_index is None or start_index < 0:
            return
        keys = outcar_lines[start_index+2].strip().split()[3:]
        charges = {}
        if self.n_ions is not None:
            for line in outcar_lines[start_index+4:start_index+4+self.n_ions]:
                index = int(line.strip().split()[0])
                charges[index] = dict([(k, float(v)) for k, v in zip(keys, line.strip().split()[1:])])
            return charges
        else:
            for line in outcar_lines[start_index+4:]:
                if line.startswith('----------------') or not line.strip():
                    break
                index = int(line.strip().split()[0])
                charges[index] = dict([(k, float(v)) for k, v in zip(keys, line.strip().split()[1:])])
            return charges

    def read_orb_projected_magnetization(self):
        """Read the orbital-projected magnetization block (example below) in a VASP OUTCAR file, if present.

         magnetization (x)

        # of ion       s       p       d       tot
        ------------------------------------------
            1       -0.000   0.018   0.033   0.050
            2       -0.000   0.018   0.034   0.052
            3       -0.003   0.004  -0.009  -0.008
            4       -0.002   0.004  -0.016  -0.014
            5       -0.008   0.001  -0.187  -0.195
            6       -0.008   0.000  -0.187  -0.196
            ...
           11        0.008   0.031   1.069   1.108
           12        0.009   0.030   1.058   1.097
        --------------------------------------------------
        tot         -0.005   0.147   1.792   1.934

        :return: {atom1: {'s': magmom_s1, 'p': magmom_p1, ...}, atom2: ...}
        :rtype: dict(int, dict(str, float))
        """
        if self.outcar is None:
            return
        outcar_lines = self.outcar.splitlines()
        start_index = None
        for r_index, line in enumerate(outcar_lines[::-1]):
            if 'magnetization (x)' in line:
                if '# of ion' in outcar_lines[len(outcar_lines) - r_index + 1]:
                    start_index = len(outcar_lines) - r_index - 1
                    break
        if start_index is None or start_index < 0:
            return
        keys = outcar_lines[start_index+2].strip().split()[3:]
        moments = {}
        if self.n_ions is not None:
            for line in outcar_lines[start_index+4:start_index+4+self.n_ions]:
                index = int(line.strip().split()[0])
                moments[index] = dict([(k, float(v)) for k, v in zip(keys, line.strip().split()[1:])])
            return moments
        else:
            for line in outcar_lines[start_index+4:]:
                if line.startswith('----------------') or not line.strip():
                    break
                index = int(line.strip().split()[0])
                moments[index] = dict([(k, float(v)) for k, v in zip(keys, line.strip().split()[1:])])
            return moments



import os
import json

# read in VASP INCAR tags and the corresponding groups
current_dir = os.path.dirname(os.path.abspath(__file__))
incar_tags_file = os.path.join(current_dir, 'incar_tags_dict.json')
# module variable VASP_INCAR_TAGS
with open(incar_tags_file, 'r') as fr:
    VASP_INCAR_TAGS = json.load(fr)

# read in default VASP settings for difference calculation_types
calculation_types = ['relaxation', 'static', 'acc_std_relax', 'sc_forces']
# module variable DEFAULT_VASP_INCAR_SETTINGS
DEFAULT_VASP_INCAR_SETTINGS = {}
for calculation_type in calculation_types:
    settings_file = os.path.join(current_dir, '{}.json'.format(calculation_type))
    with open(settings_file, 'r') as fr:
        DEFAULT_VASP_INCAR_SETTINGS[calculation_type] = json.load(fr)

# read in Hubbard U values (currently only from Wang et al., PRB 73, 195107 (2006))
hubbard_values_file = os.path.join(current_dir, 'hubbards.json')
with open(hubbard_values_file, 'r') as fr:
    HUBBARD_U_VALUES = json.load(fr)



import os
import unittest
from kelpie.vasp_input_generator import VaspInputGenerator, VaspInputError


sample_vasp_input_dir = os.path.join(os.path.dirname(__file__), 'sample_vasp_input')


class TestVaspInputGenerator(unittest.TestCase):
    """Base class to test kelpie.vasp_input_generator.VaspInputGenerator class."""

    def test_structure_setter(self):
        from kelpie import io
        s = io.read_poscar(os.path.join(sample_vasp_input_dir, 'POSCAR.all_OK'))
        ig = VaspInputGenerator(structure=s)
        self.assertEqual(ig.structure, s)

    ig = VaspInputGenerator(structure=os.path.join(sample_vasp_input_dir, 'POSCAR.all_OK'))

    def test_structure_setter_error(self):
        with self.assertRaises(VaspInputError):
            VaspInputGenerator(structure='non_existent_POSCAR')

    def test_calculation_settings_setter_defaults(self):
        self.assertTrue(self.ig.calculation_settings['lcharg'])
        self.assertAlmostEqual(self.ig.calculation_settings['potim'], 0.187)

    def test_calculation_settings_setter_encut_ediff(self):
        self.assertAlmostEqual(self.ig.calculation_settings['encut'], 600.)
        self.assertAlmostEqual(self.ig.calculation_settings['ediff'], 15E-08)

    def test_write_location_setter_default(self):
        self.assertEqual(self.ig.write_location, os.path.abspath(os.path.dirname(__file__)))

    def test_tag_value_formatter(self):
        self.assertEqual(self.ig.vasp_tag_value_formatter([27, 'acc', 0.15, True]), '27 ACC 1.50E-01 .TRUE.')

    def test_format_vasp_tag(self):
        self.assertEqual(self.ig.format_vasp_tag('ediff', 1E-6), 'EDIFF          = 1.00E-06')

    @unittest.skip
    def test_get_vasp_potcar_normal(self):
        self.assertIsInstance(self.ig.POTCAR, str)
        with open(os.path.join(sample_vasp_input_dir, 'POTCAR.LiMnO'), 'r') as fr:
            self.assertEqual(self.ig.POTCAR, fr.read())

    @unittest.skip
    def test_get_vasp_potcar_different_label(self):
        from kelpie.vasp_settings.incar import DEFAULT_VASP_INCAR_SETTINGS
        calc_sett = DEFAULT_VASP_INCAR_SETTINGS['relaxation']
        calc_sett['potcar_settings'].update({'element_potcars': {'Mn1': 'Mn_pv', 'Mn2': 'Mn'}})
        ig = VaspInputGenerator(structure=os.path.join(sample_vasp_input_dir, 'POSCAR.structure_OK'),
                                calculation_settings=calc_sett)
        with open(os.path.join(sample_vasp_input_dir, 'Li_sv__Mn_pv__Mn__O.POTCAR')) as fr:
            self.assertEqual(ig.POTCAR, fr.read())

    @unittest.skip
    def test_get_highest_enmax(self):
        self.assertAlmostEqual(self.ig.get_highest_enmax(self.ig.POTCAR), 499.034)

    def test_roundup_encut(self):
        self.assertEqual(VaspInputGenerator.roundup_encut(271.335), 280)

    def test_set_calculation_encut(self):
        self.assertEqual(self.ig.calculation_settings['encut'], 600)

    def test_scale_ediff_per_atom(self):
        self.assertAlmostEqual(self.ig.calculation_settings.get('ediff'), 15E-8)

    def test_get_vasp_incar(self):
        self.assertEqual(self.ig.INCAR.split('\n')[0], '### general ###')

    @unittest.skip
    def test_write_vasp_input_files(self):
        import shutil
        self.ig.write_vasp_input_files()
        self.assertTrue(os.path.isdir('Li6Mn2O7'))
        self.assertTrue(os.path.isfile('Li6Mn2O7/POSCAR'))
        self.assertTrue(os.path.isfile('Li6Mn2O7/POTCAR'))
        self.assertTrue(os.path.isfile('Li6Mn2O7/INCAR'))
        shutil.rmtree('Li6Mn2O7')


if __name__ == '__main__':
    unittest.main()



import os
import unittest


sample_vasp_output_dir = os.path.join(os.path.dirname(__file__), 'sample_vasp_output')

class TestDOS(unittest.TestCase):
    """Base class to test `kelpie.dos.DOS`."""

    from kelpie import dos
    relaxation_dos = dos.DOS(vasprunxml_file=os.path.join(sample_vasp_output_dir, 'relaxation_vasprun.xml.gz'))
    static_dos = dos.DOS(vasprunxml_file=os.path.join(sample_vasp_output_dir, 'static_vasprun.xml.gz'))

    def test_fermi_energy(self):
        self.assertAlmostEqual(self.relaxation_dos.fermi_energy, -1.1180374)
        self.assertAlmostEqual(self.static_dos.fermi_energy, -2.23588190)

    def test_energies(self):
        self.assertEqual(len(self.relaxation_dos.energies), 301)
        self.assertEqual(len(self.static_dos.energies), 3000)
        self.assertAlmostEqual(self.relaxation_dos.energies[4], -37.8519626)
        self.assertAlmostEqual(self.static_dos.energies[2], -21.77711810)

    def test_total_dos(self):
        self.assertListEqual(list(self.relaxation_dos.total_dos.keys()), ['spin_1', 'spin_2'])
        self.assertListEqual(list(self.static_dos.total_dos.keys()), ['spin_1'])
        self.assertEqual(len(self.relaxation_dos.total_dos['spin_1']), 301)
        self.assertEqual(len(self.static_dos.total_dos['spin_1']), 3000)
        self.assertAlmostEqual(self.relaxation_dos.total_dos['spin_2'][200], 3.8714)
        self.assertAlmostEqual(self.static_dos.total_dos['spin_1'][2000], 0.1536)

    def test_total_integrated_dos(self):
        self.assertEqual(len(self.relaxation_dos.total_integrated_dos['spin_1']), 301)
        self.assertEqual(len(self.static_dos.total_integrated_dos['spin_1']), 3000)
        self.assertAlmostEqual(self.relaxation_dos.total_integrated_dos['spin_2'][200], 7.7644)
        self.assertAlmostEqual(self.static_dos.total_integrated_dos['spin_1'][2000], 14.1469)

    def test_projected_dos(self):
        with self.assertRaises(NotImplementedError):
            self.relaxation_dos.get_projected_dos()
        with self.assertRaises(NotImplementedError):
            self.static_dos.get_projected_dos()

    def test_is_metal(self):
        self.assertListEqual(list(self.relaxation_dos.is_metal.keys()), ['spin_1', 'spin_2'])
        self.assertListEqual(list(self.static_dos.is_metal.keys()), ['spin_1'])
        self.assertTrue(self.relaxation_dos.is_metal['spin_2'])
        self.assertFalse(self.static_dos.is_metal['spin_1'])

    def test_find_vbm(self):
        self.assertIsNone(self.relaxation_dos.vbm['spin_2'])
        self.assertAlmostEqual(self.static_dos.vbm['spin_1'], -0.0166181)

    def test_find_cbm(self):
        self.assertIsNone(self.relaxation_dos.cbm['spin_1'])
        self.assertAlmostEqual(self.static_dos.cbm['spin_1'], 6.1307819)

    def test_calculate_band_gap(self):
        self.assertAlmostEqual(self.relaxation_dos.band_gap['spin_1'], 0.)
        self.assertAlmostEqual(self.relaxation_dos.band_gap['spin_2'], 0.)
        self.assertAlmostEqual(self.static_dos.band_gap['spin_1'], 6.1474)


if __name__ == '__main__':
    unittest.main()



import numpy
from collections import defaultdict
from kelpie.data import STD_ATOMIC_WEIGHTS


class AtomError(Exception):
    """Base class for error(s) in Atom objects."""
    pass


class Atom(object):
    """Base class to store an atom."""

    def __init__(self,
                 coordinates=None,
                 species=None,
                 magmom=None):
        """Constructor.

        :param coordinates: Iterable of the x, y, z coordinates of the atom (fractional or cartesian)
        :param species: Elemental species at the site.
        """
        self._coordinates = None
        self.coordinates = coordinates

        self._species = None
        self.species = species

        self._magmom = None
        self.magmom = magmom

    @property
    def coordinates(self):
        return self._coordinates

    @coordinates.setter
    def coordinates(self, coordinates):
        if coordinates is None:
            error_message = 'Atomic coordinates must be specified'
            raise AtomError(error_message)
        try:
            coord = numpy.array(coordinates, dtype='float')
        except ValueError:
            error_message = 'Could not convert atomic coordinates into float'
            raise AtomError(error_message)
        else:
            if numpy.shape(coord) != (3,):
                error_message = 'coordinates must be a 1x3-shaped iterable'
                raise AtomError(error_message)
            for i in range(3):
                if numpy.isnan(coord[i]):
                    error_message = 'Atom coordinate ({}) is None'.format(i)
                    raise AtomError(error_message)
        self._coordinates = coord.tolist()

    @property
    def species(self):
        return self._species

    @species.setter
    def species(self, species):
        if species is None:
            error_message = 'Atomic species at the site must be specified'
            raise AtomError(error_message)
        try:
            if not any([species.startswith(e) for e in STD_ATOMIC_WEIGHTS]):
                error_message = 'Species label must start with a known element symbol'
                raise AtomError(error_message)
        except AttributeError:
            error_message = 'Atomic species input must be a String'
            raise AtomError(error_message)
        else:
            self._species = species


    @property
    def magmom(self):
        return self._magmom

    @magmom.setter
    def magmom(self, magmom):
        if magmom is None:
            self._magmom = 0.
            return
        try:
            self._magmom = float(magmom)
        except (ValueError, TypeError) as error:
            error_message = 'Magnetic moment on an atom should be a decimal or integer'
            raise AtomError(error_message)

    @property
    def x(self):
        return self._coordinates[0]

    @property
    def y(self):
        return self._coordinates[1]

    @property
    def z(self):
        return self._coordinates[2]

    @property
    def element(self):
        return self._species


class StructureError(Exception):
    """Base class for error(s) in Structure objects."""
    pass


class Structure(object):
    """Base class to store a crystal structure."""

    def __init__(self,
                 scaling_constant=None,
                 lattice_vectors=None,
                 coordinate_system=None,
                 atoms=None,
                 comment=None):
        """Constructor.

        :param scaling_constant: Float to scale all the lattice vectors with. Defaults to 1.0
        :param lattice_vectors: A 3x3 Iterable of Float with the cell vectors.
        :param atoms: List of `kelpie.Atom' objects.
        :param coordinate_system: String specifying the coordinate system ("Cartesian"/"Direct")
        """

        self._scaling_constant = None
        self.scaling_constant = scaling_constant

        self._lattice_vectors = None
        self.lattice_vectors = lattice_vectors

        self._coordinate_system = None
        self.coordinate_system = coordinate_system

        self._atoms = None
        self.atoms = atoms

        self._comment = None
        self.comment = comment

    @property
    def scaling_constant(self):
        return self._scaling_constant

    @scaling_constant.setter
    def scaling_constant(self, scaling_constant):
        if scaling_constant is None:
            self._scaling_constant = 1.0
        else:
            try:
                self._scaling_constant = float(scaling_constant)
            except ValueError:
                error_message = 'Could not convert scaling constant to float'
                raise StructureError(error_message)

    @property
    def lattice_vectors(self):
        return self._lattice_vectors

    @lattice_vectors.setter
    def lattice_vectors(self, lattice_vectors):
        if lattice_vectors is None:
            return
        try:
            lv = numpy.array(lattice_vectors, dtype='float')
        except ValueError:
            error_message = 'Could not convert lattice vector component(s) into float'
            raise StructureError(error_message)
        else:
            if numpy.shape(lv) != (3, 3):
                error_message = '`lattice_vectors` must be a 3x3-shaped iterable'
                raise StructureError(error_message)
            for i in range(3):
                for j in range(3):
                    if numpy.isnan(lv[i][j]):
                        error_message = 'Lattice vector component ({}, {}) is None'.format(i, j)
                        raise StructureError(error_message)
            self._lattice_vectors = lv.tolist()

    @staticmethod
    def get_volume(lattice_vectors=None):
        """
        Return the volume of the lattice vectors specified as their absolute determinant.
        Return None if lattice vectors are not specified.
        """
        if lattice_vectors is None:
            return
        return abs(numpy.linalg.det(lattice_vectors))

    @property
    def volume(self):
        """Volume of the structure."""
        return self.get_volume(lattice_vectors=self.lattice_vectors)

    @property
    def volume_pa(self):
        """Volume of the structure per atom."""
        if not self.atoms:
            return
        return self.volume/float(self.natoms)

    @property
    def coordinate_system(self):
        return self._coordinate_system

    @coordinate_system.setter
    def coordinate_system(self, coordinate_system):
        if coordinate_system is None:
            return
        elif coordinate_system.lower() in ['direct', 'crystal', 'fractional']:
            self._coordinate_system = 'Direct'
        elif coordinate_system.lower() in ['cartesian', 'angstrom']:
            self._coordinate_system = 'Cartesian'
        else:
            error_message = 'Coordinate system not recognized. Options: Direct/Crystal/Fractional or Cartesian/Angstrom'
            raise StructureError(error_message)

    @property
    def atoms(self):
        return self._atoms

    @atoms.setter
    def atoms(self, atoms):
        if atoms is None:
            self._atoms = []
            return
        if not all([isinstance(atom, Atom) for atom in atoms]):
            error_message = '`atoms` must be an iterable of `kelpie.Atom` objects'
            raise StructureError(error_message)
        self._atoms = atoms

    def add_atom(self, atom):
        """Add `atom` to the structure.

        :raise StructureError: if `atom` is not a `structure.Atom` object."""
        if not isinstance(atom, Atom):
            error_message = '`atom` must be a `kelpie.structure.Atom` object'
            raise StructureError(error_message)
        self._atoms.append(atom)

    @property
    def natoms(self):
        return len(self.atoms)

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, comment):
        if comment is None:
            comment = 'Comment'
        self._comment = comment

    @property
    def composition_dict(self):
        if self.atoms is None:
            return {}
        composition_dict = defaultdict(int)
        for atom in self.atoms:
            composition_dict[atom.species] += 1
        return composition_dict

    @property
    def list_of_species(self):
        if self.atoms is None:
            return []
        return sorted(set([atom.species for atom in self.atoms]))

    @property
    def structural_formula(self):
        return ''.join(['{}{}'.format(e, self.composition_dict[e]) for e in self.list_of_species])

    def _get_magmom_tag(self):
        if 'A' in self._missing_structure_components():
            error_message = 'The structure has no atoms to generate the MAGMOM tag from'
            raise StructureError(error_message)
        magmoms = []
        for species in self.list_of_species:
            for atom in self.atoms:
                if atom.species == species:
                    magmoms.append(atom.magmom)
        magmom_str = []
        m_count = 1
        for i in range(len(magmoms))[1:]:
            if abs(magmoms[i] - magmoms[i-1]) < 1E-3:
                m_count += 1
            else:
                magmom_str.append('{}*{:.3f}'.format(m_count, magmoms[i-1]))
                m_count = 1
        magmom_str.append('{}*{:.3f}'.format(m_count, magmoms[-1]))
        return ' '.join(magmom_str)

    @property
    def MAGMOM(self):
        return self._get_magmom_tag()

    def _get_vasp_poscar(self):
        """Construct the VASP POSCAR.

        :return: String with the contents of a VASP 5 POSCAR file
        """
        self.check_structure_is_complete()
        poscar = []
        # system title
        poscar.append(self.comment)
        # scaling factor
        poscar.append('{:18.14f}'.format(self.scaling_constant))
        # lattice_vectors
        for lv in self.lattice_vectors:
            poscar.append('  '.join(['{:>18.14f}'.format(_lv) for _lv in lv]))
        # list of species
        poscar.append(' '.join(['{:>4s}'.format(species) for species in self.list_of_species]))
        # list of number of atoms of each species
        poscar.append(' '.join(['{:>4d}'.format(self.composition_dict[e]) for e in self.list_of_species]))
        # coordinate system
        poscar.append('{}'.format(self.coordinate_system))
        # list of atomic coordinates
        for species in self.list_of_species:
            for atom in self.atoms:
                if atom.species == species:
                    poscar.append('  '.join(['{:>18.14f}'.format(ac) for ac in atom.coordinates]))
        return '\n'.join(poscar)

    @property
    def POSCAR(self):
        """Structure in the VASP 5 POSCAR format."""
        return self._get_vasp_poscar()

    def check_structure_is_complete(self):
        """Check if structure has all components required for a DFT calculation.

        :raise StructureError: if any necessary component is missing.
        """
        missing = self._missing_structure_components()
        if missing:
            error_message = ["Structure incomplete. Following components are missing:"]
            for m in missing:
                error_message.append(self._missing_components_message(m))
            raise StructureError('\n'.join(error_message))

    def _missing_structure_components(self):
        missing = []
        if not self.lattice_vectors:
            missing.append('L')
        if not self.atoms:
            missing.append('A')
        if not self.coordinate_system:
            missing.append('C')
        return missing

    @staticmethod
    def _missing_components_message(m):
        m_dict = {'L': 'Lattice vectors',
                  'A': 'Atoms',
                  'C': 'Coordinate system'}
        return '[{}]: {}'.format(m, m_dict[m])




import os
import unittest
from kelpie import io


sample_vasp_input_dir = os.path.join(os.path.dirname(__file__), 'sample_vasp_input')


def _poscar_lines(poscar_file):
    with open(poscar_file, 'r') as fr:
        poscar_lines = [line.strip() for line in fr.readlines()]
    return poscar_lines


class TestIOReadPOSCAR(unittest.TestCase):
    """Base class to test `io.read_poscar()`"""

    def setUp(self):
        self.poscar_OK = os.path.join(sample_vasp_input_dir, 'POSCAR.structure_OK')

    def test_poscar_file_not_found(self):
        with self.assertRaises(FileNotFoundError):
            io.read_poscar(poscar_file='POSCAR.file_not_found')

    def test_consistent_number_of_atoms(self):
        self.assertTrue(io._consistent_number_of_atoms(_poscar_lines(self.poscar_OK)))
        error_poscar = os.path.join(sample_vasp_input_dir, 'POSCAR.consistent_number_of_atoms_error')
        self.assertFalse(io._consistent_number_of_atoms(_poscar_lines(error_poscar)))

    def test_system_title(self):
        self.assertEqual(io._system_title(_poscar_lines(self.poscar_OK)), 'Fe-Li-O-P')

    def test_scaling_constant(self):
        self.assertEqual(io._scaling_constant(_poscar_lines(self.poscar_OK)), 1.0)
        error_poscar = os.path.join(sample_vasp_input_dir, 'POSCAR.scaling_factor_error')
        with self.assertRaises(io.KelpieIOError):
            io._scaling_constant(_poscar_lines(error_poscar))

    def test_lattice_vectors(self):
        lattice_vectors = io._lattice_vectors(_poscar_lines(self.poscar_OK))
        self.assertIsInstance(lattice_vectors, list)
        self.assertEqual(len(lattice_vectors), 3)
        self.assertEqual(len(lattice_vectors[2]), 3)
        self.assertAlmostEqual(lattice_vectors[0], [-4.1048998, -2.0524499, 2.0524499])
        error_poscar = os.path.join(sample_vasp_input_dir, 'POSCAR.lattice_vectors_error')
        with self.assertRaises(io.KelpieIOError):
            io._lattice_vectors(_poscar_lines(error_poscar))

    def test_list_of_species(self):
        self.assertIsInstance(io._list_of_species(_poscar_lines(self.poscar_OK)), list)
        self.assertEqual(io._list_of_species(_poscar_lines(self.poscar_OK)), ['Li', 'Mn1', 'Mn2', 'O'])
        error_poscar = os.path.join(sample_vasp_input_dir, 'POSCAR.list_of_species_error')
        with self.assertRaises(io.KelpieIOError):
            io._list_of_species(_poscar_lines(error_poscar))

    def test_list_of_number_of_atoms(self):
        self.assertIsInstance(io._list_of_number_of_atoms(_poscar_lines(self.poscar_OK)), list)
        self.assertEqual(io._list_of_number_of_atoms(_poscar_lines(self.poscar_OK)), [6, 1, 1, 7])
        error_poscar = os.path.join(sample_vasp_input_dir, 'POSCAR.list_of_number_of_atoms_error')
        with self.assertRaises(io.KelpieIOError):
            io._list_of_number_of_atoms(_poscar_lines(error_poscar))

    def test_repeating_list_of_species(self):
        rl = ['Li']*6 + ['Mn1'] + ['Mn2'] + ['O']*7
        self.assertEqual(io._repeating_list_of_species(_poscar_lines(self.poscar_OK)), rl)

    def test_coordinate_system(self):
        self.assertEqual(io._coordinate_system(_poscar_lines(self.poscar_OK)), 'Direct')
        error_poscar = os.path.join(sample_vasp_input_dir, 'POSCAR.selective_dynamics_error')
        with self.assertRaises(NotImplementedError):
            io._coordinate_system(_poscar_lines(error_poscar))
        error_poscar = os.path.join(sample_vasp_input_dir, 'POSCAR.coordinate_system_error')
        with self.assertRaises(io.KelpieIOError):
            io._coordinate_system(_poscar_lines(error_poscar))

    def test_list_of_atomic_coordinates(self):
        list_of_atomic_coordinates = io._list_of_atomic_coordinates(_poscar_lines(self.poscar_OK))
        self.assertEqual(len(list_of_atomic_coordinates), 15)
        self.assertEqual(len(list_of_atomic_coordinates[2]), 3)
        self.assertAlmostEqual(list_of_atomic_coordinates[14][0], 0.74999999)
        error_poscar = os.path.join(sample_vasp_input_dir, 'POSCAR.list_of_atomic_coordinates_error')
        with self.assertRaises(io.KelpieIOError):
            io._list_of_atomic_coordinates(_poscar_lines(error_poscar))


if __name__ == '__main__':
    unittest.main()




import argparse


class KelpieArgumentParser(argparse.ArgumentParser):
    """Argument parser wrapper for command line kelpie."""
    def __init__(self, *args, **kwargs):
        if not kwargs.get('formatter_class'):
            kwargs['formatter_class'] = argparse.RawTextHelpFormatter
        super(KelpieArgumentParser, self).__init__(*args, **kwargs, add_help=False)

        general_args = self.add_argument_group(title='General arguments')
        scheduler_args = self.add_argument_group(title='Batch scheduler-related arguments')
        calculation_args = self.add_argument_group(title='DFT calculation-related arguments')

        ######################################################################
        # General arguments (optional)
        ######################################################################
        input_structure_help = """Location of the file with the structure to calculate.
REQUIRED if mode is "breed" (see below).

"""
        general_args.add_argument('-i', '--input-structure-file',
                                  default=None,
                                  help=input_structure_help)

        mode_help = """Mode to run kelpie in:
breed: Generate the necessary directory structure, job file, and submit job.
graze: Perform the specified calculation workflow using the specified DFT code.
herd:  Gather data for all the calculations performed during grazing.

"""
        general_args.add_argument('-m', '--mode',
                                  default='breed',
                                  choices=['breed', 'graze', 'herd'],
                                  help=mode_help)

        submit_help = """Should the batch job be submitted to the scheduler?
Only relevant for "breed" mode, ignored otherwise. Defaults to "true".

"""
        general_args.add_argument('-s', '--submit-batch-job',
                                  default='true',
                                  choices=['true', 'false'],
                                  help=submit_help)

        run_location_help = """Path/location where the calculations need to be run.
Directories along the path specified will be created if not already present.
Defaults to the current working directory.

"""
        general_args.add_argument('-l', '--run-location',
                                  default=None,
                                  help=run_location_help)

        help_help = """Show this help message and exit

"""
        general_args.add_argument('-h', '--help',
                                  action='help',
                                  help=help_help)

        ######################################################################
        # Batch scheduler related arguments (optional)
        ######################################################################
        host_scheduler_help = """Name of a predefined host or path to a JSON file
with the scheduler settings. If a name is specified, must be one of the hosts
defined in `kelpie.scheduler_settings` (i.e., with a corresponding
"[host_scheduler].json" file. their values. Path to JSON file takes precedence.
Defaults to "cori_knl".

"""
        scheduler_args.add_argument('-hss', '--host-scheduler-settings',
                                    default=None,
                                    help=host_scheduler_help)

        custom_scheduler_help = """Path to a JSON file with a dictionary of tags
and *nondefault* values to go into the batch script. (The default settings
dictionary, if loaded based on the host-scheduler specified with the "-hs" tag
will be updated.

"""
        scheduler_args.add_argument('-css', '--custom-scheduler-settings',
                                    default=None,
                                    help=custom_scheduler_help)

        batch_script_help = """Name of a predefined template file or path to a file
with a template of the batch script to use to submit jobs. Name of the
predefined template must be one of those in the "scheduler_settings" directory.
Path to a template file takes precedence.
Defaults to "cori".

"""
        scheduler_args.add_argument('-bst', '--batch-script-template',
                                    default=None,
                                    help=batch_script_help)

        ######################################################################
        # DFT calculation/settings related arguments (optional)
        ######################################################################
        calculation_workflow_help = """Tag specifying the calculation workflow to
perform. Currently only "relaxation" and "static" implemented. Default settings
for each calculation in the workflow are predefined.
Defaults to "relaxation".

"""
        calculation_args.add_argument('-w', '--calculation-workflow',
                                      default='relaxation',
                                      choices=['relaxation', 'static',
                                               'acc_std_relax', 'sc_forces'],
                                      help=calculation_workflow_help)

        custom_calculation_settings_help = """Absolute path to a JSON file with a
dictionary of *nondefault* INCAR and POTCAR settings for each calculation type. The
default calculation settings dictionary loaded according to the workflow will be
updated. E.g., {"relaxation": {"ediff": 1E-8}, "static": {"sigma": 0.05}}

"""
        calculation_args.add_argument('-ccs', '--custom-calculation-settings',
                                      default=None,
                                      help=custom_calculation_settings_help)




import os
import glob
import json

current_dir = os.path.dirname(os.path.abspath(__file__))
host_setting_files = glob.glob(os.path.join(current_dir, '*.json'))

# module variable DEFAULT_SCHEDULER_SETTINGS
DEFAULT_SCHEDULER_SETTINGS = {}
for host_setting_file in host_setting_files:
    host_tag = os.path.splitext(os.path.basename(host_setting_file))[0]
    with open(host_setting_file, 'r') as fr:
        DEFAULT_SCHEDULER_SETTINGS[host_tag] = json.load(fr)



import os
import shutil
import datetime
from contextlib import contextmanager
from kelpie.utils.serialization import jsonable


@contextmanager
def change_working_dir(new_dir):
    current_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(current_dir)


def time_stamped_folder(base_name='archive'):
    now = datetime.datetime.now()
    return '_'.join([base_name, jsonable(now)])


def backup_files(list_of_files=None, folder_name='archive', gzip=True):
    if not list_of_files:
        list_of_files = ['POSCAR', 'INCAR', 'OSZICAR', 'OUTCAR', 'vasprun.xml']
    ts_folder = time_stamped_folder(base_name=folder_name)
    os.mkdir(ts_folder)
    for src_file in list_of_files:
        if not os.path.isfile(src_file):
            continue
        if gzip:
            dest_file = os.path.join(ts_folder, src_file+'.gz')
            with open(src_file, 'rb') as fsrc:
                with open(dest_file, 'wb') as fdest:
                    shutil.copyfileobj(fsrc, fdest)
        else:
            dest_file = os.path.join(ts_folder, src_file)
            shutil.copyfile(src_file, dest_file)


def copy_files(src_folder=None, dest_folder=None, list_of_filenames=None):
    if any([s is None for s in [src_folder, dest_folder, list_of_filenames]]):
        return
    for fn in list_of_filenames:
        src_file = os.path.join(src_folder, fn)
        dest_file = os.path.join(dest_folder, fn)
        shutil.copyfile(src_file, dest_file)




import os
import unittest
from kelpie.vasp_calculation_data import VaspCalculationData


sample_vasp_output_dir = os.path.join(os.path.dirname(__file__), 'sample_vasp_output')


class TestVaspCalculationData(unittest.TestCase):
    """Base class to test :class:`kelpie.vasp_calculation_data.VaspCalculationData`."""

    relaxation_xml = os.path.join(sample_vasp_output_dir, 'relaxation_vasprun.xml.gz')
    relaxation_vcd = VaspCalculationData(vasprunxml_file=relaxation_xml)

    static_xml = os.path.join(sample_vasp_output_dir, 'static_vasprun.xml.gz')
    static_outcar = os.path.join(sample_vasp_output_dir, 'static_OUTCAR.gz')
    static_vcd = VaspCalculationData(vasprunxml_file=static_xml, vasp_outcar_file=static_outcar)

    def test_vasprunxml_file(self):
        self.assertEqual(self.relaxation_vcd.vasprunxml_file, os.path.join(sample_vasp_output_dir,
                                                                           'relaxation_vasprun.xml.gz'))
        self.assertEqual(self.static_vcd.vasprunxml_file, os.path.join(sample_vasp_output_dir,
                                                                       'static_vasprun.xml.gz'))

    def test_vxparser(self):
        from kelpie.vasp_output_parser import VasprunXMLParser
        self.assertIsInstance(self.relaxation_vcd.vxparser, VasprunXMLParser)
        self.assertIsInstance(self.static_vcd.vxparser, VasprunXMLParser)

    def test_vasp_outcar_file(self):
        self.assertIsNone(self.relaxation_vcd.vasp_outcar_file)
        self.assertEqual(self.static_vcd.vasp_outcar_file, os.path.join(sample_vasp_output_dir, 'static_OUTCAR.gz'))

    def test_voutparser(self):
        from kelpie.vasp_output_parser import VaspOutcarParser
        self.assertIsInstance(self.relaxation_vcd.voutparser, VaspOutcarParser)
        self.assertIsNone(self.relaxation_vcd.voutparser.vasp_outcar_file)
        self.assertIsInstance(self.static_vcd.voutparser, VaspOutcarParser)

    def test_runtimestamp(self):
        from datetime import datetime
        relaxation_timestamp = datetime(year=2016, month=9, day=22, hour=20, minute=31, second=26)
        static_timestamp = datetime(year=2018, month=3, day=20, hour=1, minute=12, second=11)
        self.assertEqual(relaxation_timestamp, self.relaxation_vcd.run_timestamp)
        self.assertEqual(static_timestamp, self.static_vcd.run_timestamp)

    def test_composition_info(self):
        self.assertSetEqual(set(self.relaxation_vcd.composition_info.keys()), {'Ca', 'F'})
        self.assertEqual(self.relaxation_vcd.composition_info['F']['natoms'], 2)
        self.assertAlmostEqual(self.static_vcd.composition_info['Na']['atomic_mass'], 22.99)
        self.assertAlmostEqual(self.relaxation_vcd.composition_info['Ca']['valence'], 10.)
        self.assertEqual(self.static_vcd.composition_info['Na']['pseudopotential'], 'PAW_PBE Na_pv 19Sep2006')

    def test_unit_cell_formula(self):
        self.assertDictEqual(self.relaxation_vcd.unit_cell_formula, {'Ca': 1, 'F': 2})
        self.assertDictEqual(self.static_vcd.unit_cell_formula, {'Na': 1, 'F': 1})

    def test_list_of_atoms(self):
        self.assertListEqual(self.relaxation_vcd.list_of_atoms, ['Ca', 'F', 'F'])
        self.assertListEqual(self.static_vcd.list_of_atoms, ['F', 'Na'])

    def test_n_atoms(self):
        self.assertEqual(self.relaxation_vcd.n_atoms, 3)
        self.assertEqual(self.static_vcd.n_atoms, 2)

    def test_n_ionic_steps(self):
        self.assertEqual(self.relaxation_vcd.n_ionic_steps, 3)
        self.assertEqual(self.static_vcd.n_ionic_steps, 1)

    def test_scf_energies(self):
        self.assertSetEqual(set(self.relaxation_vcd.scf_energies.keys()), {0, 1, 2})
        self.assertAlmostEqual(self.relaxation_vcd.scf_energies[1][-1], -17.5606492)
        self.assertAlmostEqual(self.static_vcd.scf_energies[0][-2], -8.57103789)

    def test_entropies(self):
        self.assertAlmostEqual(self.relaxation_vcd.entropies[1], 0.01828511)
        self.assertAlmostEqual(self.static_vcd.entropies[0], 0.)

    def test_free_energies(self):
        self.assertAlmostEqual(self.relaxation_vcd.free_energies[2], -17.5616337)
        self.assertAlmostEqual(self.static_vcd.free_energies[0], -8.57103792)

    def test_forces(self):
        import numpy
        self.assertAlmostEqual(self.relaxation_vcd.forces[0][1][1], 0.)
        self.assertTupleEqual(numpy.shape(self.static_vcd.forces[0]), (2, 3))

    def test_stress_tensors(self):
        import numpy
        self.assertTupleEqual(numpy.shape(self.relaxation_vcd.stress_tensors[2]), (3, 3))
        self.assertAlmostEqual(self.relaxation_vcd.stress_tensors[2][1][1], -1.07296801)
        self.assertTupleEqual(numpy.shape(self.static_vcd.stress_tensors[0]), (3, 3))
        self.assertAlmostEqual(self.static_vcd.stress_tensors[0][2][2], 0.14074542)

    def test_lattice_vectors(self):
        self.assertAlmostEqual(self.relaxation_vcd.lattice_vectors[2][1][0], -1.94748670)
        self.assertAlmostEqual(self.static_vcd.lattice_vectors[0][1][1], 2.34638389)

    def test_atomic_coordinates(self):
        import numpy
        self.assertTupleEqual(numpy.shape(self.relaxation_vcd.atomic_coordinates[0]), (3, 3))
        self.assertTupleEqual(numpy.shape(self.static_vcd.atomic_coordinates[0]), (2, 3))
        self.assertAlmostEqual(self.relaxation_vcd.atomic_coordinates[0][1][2], 0.24999888)
        self.assertAlmostEqual(self.static_vcd.atomic_coordinates[0][0][1], 0.5)

    def test_cell_volumes(self):
        self.assertListEqual(list(self.relaxation_vcd.cell_volumes.keys()), [0, 1, 2])
        self.assertListEqual(list(self.static_vcd.cell_volumes.keys()), [0])
        self.assertAlmostEqual(self.relaxation_vcd.cell_volumes[1], 42.18428983)
        self.assertAlmostEqual(self.static_vcd.cell_volumes[0], 25.83611454)

    def test_kpoint_mesh(self):
        self.assertListEqual(self.relaxation_vcd.kpoint_mesh, [11, 12, 11])
        self.assertListEqual(self.static_vcd.kpoint_mesh, [16, 16, 16])

    def test_irreducible_kpoints(self):
        self.assertAlmostEqual(self.relaxation_vcd.irreducible_kpoints[4][0], 0.36363636)
        self.assertAlmostEqual(self.static_vcd.irreducible_kpoints[2][0], 0.125)

    def test_n_irreducible_kpoints(self):
        self.assertEqual(self.relaxation_vcd.n_irreducible_kpoints, 402)
        self.assertEqual(self.static_vcd.n_irreducible_kpoints, 145)

    def test_band_occupations(self):
        self.assertListEqual(list(self.relaxation_vcd.band_occupations.keys()), ['spin_1', 'spin_2'])
        self.assertListEqual(list(self.static_vcd.band_occupations.keys()), ['spin_1'])
        self.assertAlmostEqual(self.relaxation_vcd.band_occupations['spin_2'][1]['band_energy'][0], -37.1897)
        self.assertAlmostEqual(self.relaxation_vcd.band_occupations['spin_2'][1]['occupation'][9], 1.0353)
        self.assertAlmostEqual(self.static_vcd.band_occupations['spin_1'][2]['band_energy'][3], -21.1815)
        self.assertAlmostEqual(self.static_vcd.band_occupations['spin_1'][2]['occupation'][9], 0.)

    def test_n_bands(self):
        self.assertEqual(self.relaxation_vcd.n_bands, 18)
        self.assertEqual(self.static_vcd.n_bands, 11)

    def test_scf_looptimes(self):
        self.assertListEqual(list(self.relaxation_vcd.scf_looptimes.keys()), [0, 1, 2])
        self.assertListEqual(list(self.static_vcd.scf_looptimes.keys()), [0])
        self.assertAlmostEqual(self.relaxation_vcd.scf_looptimes[0][14], 2.24)
        self.assertAlmostEqual(self.static_vcd.scf_looptimes[0][10], 1.28)

    def test_average_scf_looptimes(self):
        self.assertAlmostEqual(self.relaxation_vcd.average_scf_looptimes[0], 2.55933333)
        self.assertAlmostEqual(self.static_vcd.average_scf_looptimes[0], 1.586363636)

    def test_average_scf_looptime(self):
        self.assertAlmostEqual(self.relaxation_vcd.average_scf_looptime, 2.7169629629)
        self.assertAlmostEqual(self.static_vcd.average_scf_looptime, 1.58636363)

    def test_average_n_scf_steps_per_ionic_step(self):
        self.assertAlmostEqual(self.relaxation_vcd.average_n_scf_steps_per_ionic_step, 11.333333333)
        self.assertAlmostEqual(self.static_vcd.average_n_scf_steps_per_ionic_step, 11.)

    def test_total_runtime(self):
        self.assertAlmostEqual(self.relaxation_vcd.total_runtime, 91.6)
        self.assertAlmostEqual(self.static_vcd.total_runtime, 17.45)

    def test_density_of_states(self):
        from kelpie.dos import DOS
        self.assertIsInstance(self.relaxation_vcd.density_of_states, DOS)
        self.assertIsInstance(self.static_vcd.density_of_states, DOS)

    def test_fermi_energy(self):
        self.assertAlmostEqual(self.relaxation_vcd.fermi_energy, -1.11803743)
        self.assertAlmostEqual(self.static_vcd.fermi_energy, -2.23588190)

    def test_dos_energy_grid(self):
        self.assertEqual(len(self.relaxation_vcd.dos_energy_grid), 301)
        self.assertEqual(len(self.static_vcd.dos_energy_grid), 3000)
        self.assertAlmostEqual(self.relaxation_vcd.dos_energy_grid[0], -38.59096257)
        self.assertAlmostEqual(self.static_vcd.dos_energy_grid[1], -21.7919181)

    def test_total_dos(self):
        self.assertListEqual(list(self.relaxation_vcd.total_dos.keys()), ['spin_1', 'spin_2'])
        self.assertListEqual(list(self.static_vcd.total_dos.keys()), ['spin_1'])
        self.assertEqual(len(self.relaxation_vcd.total_dos['spin_2']), 301)
        self.assertEqual(len(self.static_vcd.total_dos['spin_1']), 3000)
        self.assertAlmostEqual(self.relaxation_vcd.total_dos['spin_2'][99], -0.0836)
        self.assertAlmostEqual(self.static_vcd.total_dos['spin_1'][2000], 0.1536)

    def test_total_integrated_dos(self):
        self.assertListEqual(list(self.relaxation_vcd.total_integrated_dos.keys()), ['spin_1', 'spin_2'])
        self.assertListEqual(list(self.static_vcd.total_integrated_dos.keys()), ['spin_1'])
        self.assertEqual(len(self.relaxation_vcd.total_integrated_dos['spin_2']), 301)
        self.assertEqual(len(self.static_vcd.total_integrated_dos['spin_1']), 3000)
        self.assertAlmostEqual(self.relaxation_vcd.total_integrated_dos['spin_2'][99], 0.9830)
        self.assertAlmostEqual(self.static_vcd.total_integrated_dos['spin_1'][2000], 14.1469)

    def test_is_metal(self):
        self.assertTrue(self.relaxation_vcd.is_metal['spin_2'])
        self.assertFalse(self.static_vcd.is_metal['spin_1'])

    def test_valence_band_maximum(self):
        self.assertIsNone(self.relaxation_vcd.valence_band_maximum['spin_1'])
        self.assertAlmostEqual(self.static_vcd.valence_band_maximum['spin_1'], -0.0166181)

    def test_conduction_band_minimum(self):
        self.assertIsNone(self.relaxation_vcd.conduction_band_minimum['spin_1'])
        self.assertAlmostEqual(self.static_vcd.conduction_band_minimum['spin_1'], 6.1307819)

    def test_dos_band_gap(self):
        self.assertAlmostEqual(self.relaxation_vcd.dos_band_gap['spin_2'], 0.)
        self.assertAlmostEqual(self.static_vcd.dos_band_gap['spin_1'], 6.1474)

    def test_orb_projected_charge(self):
        self.assertIsNone(self.relaxation_vcd.orb_projected_charge)
        self.assertListEqual(list(self.static_vcd.orb_projected_charge.keys()), list(range(1, 13)))
        self.assertListEqual(list(self.static_vcd.orb_projected_charge[1].keys()), ['s', 'p', 'd', 'tot'])
        self.assertAlmostEqual(self.static_vcd.orb_projected_charge[1]['s'], 0.476)
        self.assertAlmostEqual(self.static_vcd.orb_projected_charge[3]['p'], 6.110)
        self.assertAlmostEqual(self.static_vcd.orb_projected_charge[7]['d'], 0.085)
        self.assertAlmostEqual(self.static_vcd.orb_projected_charge[12]['tot'], 7.485)

    def test_orb_projected_magnetization(self):
        self.assertIsNone(self.relaxation_vcd.orb_projected_magnetization)
        self.assertListEqual(list(self.static_vcd.orb_projected_magnetization.keys()), list(range(1, 13)))
        self.assertListEqual(list(self.static_vcd.orb_projected_magnetization[1].keys()), ['s', 'p', 'd', 'tot'])
        self.assertAlmostEqual(self.static_vcd.orb_projected_magnetization[3]['s'], -0.003)
        self.assertAlmostEqual(self.static_vcd.orb_projected_magnetization[1]['p'], 0.018)
        self.assertAlmostEqual(self.static_vcd.orb_projected_magnetization[11]['d'], 1.069)
        self.assertAlmostEqual(self.static_vcd.orb_projected_magnetization[12]['tot'], 1.097)

    def test_total_orb_projected_charge(self):
        self.assertIsNone(self.relaxation_vcd.total_orb_projected_charge)
        self.assertListEqual(list(self.static_vcd.total_orb_projected_charge.keys()), ['s', 'p', 'd', 'tot'])
        self.assertAlmostEqual(self.static_vcd.total_orb_projected_charge['s'], 14.948)
        self.assertAlmostEqual(self.static_vcd.total_orb_projected_charge['p'], 32.175)
        self.assertAlmostEqual(self.static_vcd.total_orb_projected_charge['d'], 35.923)
        self.assertAlmostEqual(self.static_vcd.total_orb_projected_charge['tot'], 83.047)

    def test_total_orb_projected_magnetization(self):
        self.assertIsNone(self.relaxation_vcd.total_orb_projected_magnetization)
        self.assertListEqual(list(self.static_vcd.total_orb_projected_magnetization.keys()), ['s', 'p', 'd', 'tot'])
        self.assertAlmostEqual(self.static_vcd.total_orb_projected_magnetization['s'], -0.004)
        self.assertAlmostEqual(self.static_vcd.total_orb_projected_magnetization['p'], 0.148)
        self.assertAlmostEqual(self.static_vcd.total_orb_projected_magnetization['d'], 1.793)
        self.assertAlmostEqual(self.static_vcd.total_orb_projected_magnetization['tot'], 1.935)

    def test_initial_entropy(self):
        self.assertAlmostEqual(self.relaxation_vcd.initial_entropy, 0.01621571)
        self.assertAlmostEqual(self.static_vcd.initial_entropy, 0.)

    def test_final_entropy(self):
        self.assertAlmostEqual(self.relaxation_vcd.final_entropy, 0.01803879)
        self.assertAlmostEqual(self.static_vcd.final_entropy, 0.)

    def test_initial_free_energy(self):
        self.assertAlmostEqual(self.relaxation_vcd.initial_free_energy, -17.50045148)
        self.assertAlmostEqual(self.static_vcd.initial_free_energy, -8.57103792)

    def test_final_free_energy(self):
        self.assertAlmostEqual(self.relaxation_vcd.final_free_energy, -17.56163366)
        self.assertAlmostEqual(self.static_vcd.final_free_energy, -8.57103792)

    def test_initial_forces(self):
        self.assertAlmostEqual(self.relaxation_vcd.initial_forces[1][1], 0.)
        self.assertAlmostEqual(self.static_vcd.initial_forces[1][2], 0.)

    def test_final_forces(self):
        self.assertAlmostEqual(self.relaxation_vcd.final_forces[2][0], 0.)
        self.assertAlmostEqual(self.static_vcd.final_forces[1][0], 0.)

    def test_initial_stress_tensor(self):
        self.assertAlmostEqual(self.relaxation_vcd.initial_stress_tensor[1][1], 70.96233651)
        self.assertAlmostEqual(self.static_vcd.initial_stress_tensor[2][2], 0.14074542)

    def test_final_stress_tensor(self):
        self.assertAlmostEqual(self.relaxation_vcd.final_stress_tensor[1][1], -1.07296801)
        self.assertAlmostEqual(self.static_vcd.final_stress_tensor[2][2], 0.14074542)

    def test_initial_cell_volume(self):
        self.assertAlmostEqual(self.relaxation_vcd.initial_cell_volume, 38.86638710)
        self.assertAlmostEqual(self.static_vcd.initial_cell_volume, 25.83611454)

    def test_final_cell_volume(self):
        self.assertAlmostEqual(self.relaxation_vcd.final_cell_volume, 41.78289124)
        self.assertAlmostEqual(self.static_vcd.final_cell_volume, 25.83611454)

    def test_initial_volume_pa(self):
        self.assertAlmostEqual(self.relaxation_vcd.initial_volume_pa, 12.9554623666)
        self.assertAlmostEqual(self.static_vcd.initial_volume_pa, 12.91805727)

    def test_final_volume_pa(self):
        self.assertAlmostEqual(self.relaxation_vcd.final_volume_pa, 13.927630413)
        self.assertAlmostEqual(self.static_vcd.final_volume_pa, 12.91805727)

    def test_initial_lattice_vectors(self):
        self.assertAlmostEqual(self.relaxation_vcd.initial_lattice_vectors[2][1], -2.19517474)
        self.assertAlmostEqual(self.relaxation_vcd.initial_lattice_vectors[2][2], 3.10444588)
        self.assertAlmostEqual(self.static_vcd.initial_lattice_vectors[1][1], 2.34638389)
        self.assertAlmostEqual(self.static_vcd.initial_lattice_vectors[2][2], 2.34638389)

    def test_final_lattice_vectors(self):
        self.assertAlmostEqual(self.relaxation_vcd.final_lattice_vectors[2][1], -2.24876394)
        self.assertAlmostEqual(self.relaxation_vcd.final_lattice_vectors[2][2], 3.18023246)
        self.assertAlmostEqual(self.static_vcd.final_lattice_vectors[1][1], 2.34638389)
        self.assertAlmostEqual(self.static_vcd.final_lattice_vectors[2][2], 2.34638389)

    def test_initial_atomic_coordinates(self):
        self.assertAlmostEqual(self.relaxation_vcd.initial_atomic_coordinates[2][2], 0.74999888)
        self.assertAlmostEqual(self.static_vcd.initial_atomic_coordinates[0][1], 0.5)

    def test_final_atomic_coordinates(self):
        self.assertAlmostEqual(self.relaxation_vcd.final_atomic_coordinates[2][1], 0.49999888)
        self.assertAlmostEqual(self.static_vcd.final_atomic_coordinates[0][1], 0.5)

    def test_is_scf_converged(self):
        self.assertTrue(self.relaxation_vcd.is_scf_converged())
        self.assertFalse(self.relaxation_vcd.is_scf_converged(threshold=1E-9, each_ionic_step=True))
        self.assertTrue(self.static_vcd.is_scf_converged())
        self.assertFalse(self.static_vcd.is_scf_converged(threshold=1E-9, each_ionic_step=True))

    def test_are_forces_converged(self):
        self.assertTrue(self.relaxation_vcd.are_forces_converged())
        self.assertTrue(self.relaxation_vcd.are_forces_converged(threshold=1E-6))
        self.assertTrue(self.static_vcd.are_forces_converged())
        self.assertTrue(self.static_vcd.are_forces_converged(threshold=1E-6))

    def test_is_number_of_bands_converged(self):
        self.assertTrue(self.relaxation_vcd.is_number_of_bands_converged())
        self.assertTrue(self.static_vcd.is_number_of_bands_converged())

    def test_is_basis_converged(self):
        self.assertFalse(self.relaxation_vcd.is_basis_converged())
        self.assertTrue(self.relaxation_vcd.is_basis_converged(threshold=5E-2))
        self.assertFalse(self.relaxation_vcd.is_basis_converged(volume_only=True))
        self.assertTrue(self.relaxation_vcd.is_basis_converged(volume_only=True, threshold=3E-2))
        self.assertTrue(self.static_vcd.is_basis_converged())

    def test_is_fully_converged(self):
        self.assertFalse(self.relaxation_vcd.is_fully_converged())
        self.assertTrue(self.relaxation_vcd.is_fully_converged(scf_thresh=1E-6, each_ionic_step=True,
                                                               force_thresh=1E-6, basis_thresh=3E-2))
        self.assertTrue(self.static_vcd.is_fully_converged())


if __name__ == '__main__':
    unittest.main()



import os
import json

# read in list of VASP recommended potentials for each element
current_dir = os.path.dirname(os.path.abspath(__file__))
reco_potcars_file = os.path.join(current_dir, 'vasp_reco_potcars.json')
# module variable VASP_RECO_POTCARS
with open(reco_potcars_file, 'r') as fr:
    VASP_RECO_POTCARS = json.load(fr)




import numpy
import six
import datetime


class JSONableError(Exception):
    """Base class to handle errors associated with JSON-serializing data."""
    pass


def datetime_to_str(time):
    return '{:0>4d}{:0>2d}{:0>2d}{:0>2d}{:0>2d}'.format(time.year, time.month, time.day, time.hour, time.minute)


def jsonable(data, ignore_failures=True):
    if data is None:
        return
    elif isinstance(data, (bool, int, float, six.string_types)):
        return data
    elif isinstance(data, list):
        return [jsonable(v) for v in data]
    elif isinstance(data, datetime.datetime):
        return datetime_to_str(data)
    elif isinstance(data, numpy.ndarray):
        return jsonable(data.tolist())
    elif isinstance(data, (numpy.integer, numpy.floating, numpy.bool_)):
        return numpy.asscalar(data)
    elif isinstance(data, dict):
        json_data = {}
        for k, v in data.items():
            json_data[k] = jsonable(v)
        return json_data
    else:
        if ignore_failures:
            return
        else:
            error_message = 'Cannot serialize data type {}'.format(type(data))
            raise JSONableError(error_message)




from setuptools import setup, find_packages
import os
from kelpie import __version__


with open(os.path.join(os.path.dirname(__file__), 'README.md'), 'r') as fr:
    long_description = fr.read()


setup(
    name='kelpie',
    version=__version__,
    description='A lean Python module for cluster-side management of VASP runs',
    long_description=long_description,
    url='https://gitlab.com/hegdevinayi/kelpie',
    author='Vinay Hegde',
    author_email='hegdevinayi@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='VASP DFT supercomputer high-throughput',
    packages=find_packages(exclude=['tests*', 'docs']),
    install_requires=[
        'six',
        'lxml',
        'numpy'
    ],
    include_package_data=True,
    scripts=[
        'bin/kelpie',
    ]
)



import os
import glob

current_dir = os.path.dirname(os.path.abspath(__file__))
templates = glob.glob(os.path.join(current_dir, '*.q'))

# module variable DEFAULT_SCHEDULER_SETTINGS
SCHEDULER_TEMPLATES = {}
for template in templates:
    template_tag = os.path.splitext(os.path.basename(template))[0]
    SCHEDULER_TEMPLATES[template_tag] = os.path.abspath(template)
